"""Tools for plotting spectra."""

from __future__ import print_function
import matplotlib.pyplot as plt
import numpy as np
from uncertainties import unumpy


class PlottingError(Exception):
    """General exception for plotting.py"""

    pass


class SpectrumPlotter(object):
    """Class for handling spectrum plotting."""

    def __init__(self, spec, *fmt, **kwargs):
        """
        Args:
          spec:   Spectrum instance to plot
          fmt:    matplotlib like plot format string
          xmode:  define what is plotted on x axis ('energy' or 'channel')
          ymode:  define what is plotted on y axis ('counts', 'cps', 'cpskev')
          xlim:   set x axes limits, if set to 'default' use special scales
          ylim:   set y axes limits, if set to 'default' use special scales
          ax:     matplotlib axes object, if not provided one is created using
          yscale: matplotlib scale: 'linear', 'log', 'logit', 'symlog'
          title:  costum plot title, default is filename if available
          xlabel: costum xlabel value
          ylabel: costum ylabel value
          kwargs: arguments that are directly passed to matplotlib's plot command.
                  In addition it is possible to pass linthreshy if ylim='default'
                  and ymode='symlog'
        """

        #TODO Marco: maybe we should use tuplets for all the x/y stuff in order
        #to reduce the number of attributes (pylint ask for 7 we have 23)!
        self._xedges = None
        self._ydata = None
        self._xmode = None
        self._ymode = None
        self._xlabel = None
        self._ylabel = None
        self._ax = None
        self._xlim = None
        self._ylim = None
        self._linthreshy = None

        xmode = None
        ymode = None
        xlim = None
        ylim = None
        ax = None
        yscale = None
        title = None
        xlabel = None
        ylabel = None

        if 'xmode' in kwargs:
            xmode = kwargs.pop('xmode')
        if 'ymode' in kwargs:
            ymode = kwargs.pop('ymode')
        if 'xlim' in kwargs:
            xlim = kwargs.pop('xlim')
        if 'ylim' in kwargs:
            ylim = kwargs.pop('ylim')
        if 'ax' in kwargs:
            ax = kwargs.pop('ax')
        if 'yscale' in kwargs:
            yscale = kwargs.pop('yscale')
        if 'title' in kwargs:
            title = kwargs.pop('title')
        if 'xlabel' in kwargs:
            xlabel = kwargs.pop('xlabel')
        if 'ylabel' in kwargs:
            ylabel = kwargs.pop('ylabel')

        self.spec = spec

        if hasattr(fmt, '__len__') and len(fmt) in [0, 1]:
            self.fmt = fmt
        else:
            raise PlottingError("Wrong number of positional arguments")

        if 'linthreshy' in kwargs:
            self._linthreshy = kwargs.pop("linthreshy")

        self.xlim = xlim
        self.ylim = ylim

        self.yscale = yscale
        self.ax = ax
        self.title = title
        self.kwargs = kwargs

        self.xmode = xmode
        self.ymode = ymode

        self.xlabel = xlabel
        self.ylabel = ylabel


    @property
    def xmode(self):
        """
        Returns the current x axis plotting mode
        """

        return self._xmode


    @xmode.setter
    def xmode(self, mode):
        """
        Define x data mode, handles all data errors, requires spec.
        Defines also xedges and xlabel.

        Args:
          mode: energy (or kev), channel (or channels, chn, chns)
        """

        if mode is None:
            if self.spec.is_calibrated:
                self._xmode = 'energy'
            else:
                self._xmode = 'channel'
        else:
            if mode.lower() in ('kev', 'energy'):
                if not self.spec.is_calibrated:
                    raise PlottingError('Spectrum is not calibrated, however'
                                        ' x axis was requested as energy')
                self._xmode = 'energy'
            elif mode.lower() in ('channel', 'channels', 'chn', 'chns'):
                self._xmode = 'channel'
            else:
                raise PlottingError('Unknown x data mode: {}'.format(mode))

        if self._xmode == 'energy':
            self._xedges = self.spec.bin_edges_kev
            if self._xlabel == 'Channel' or self._xlabel is None:
                self._xlabel = 'Energy [keV]'
        elif self._xmode == 'channel':
            self._xedges = self.spec.bin_edges_raw
            if self._xlabel == 'Energy [keV]' or self._xlabel is None:
                self._xlabel = 'Channel'


    @property
    def ymode(self):
        """
        Returns the current y axis plotting mode.
        """
        return self._ymode


    @ymode.setter
    def ymode(self, mode):
        """
        Define y data mode, handles all data errors, requires spec.
        Defines also ydata and ylabel. Does not check if cps is defined.
        If it is not defined it will trow a SpectrumError.

        Args:
          mode: counts, cps, cpskev
        """

        if mode is None:
            if self.spec._counts is not None:
                self._ymode = 'counts'
            else:
                self._ymode = 'cps'
        elif mode.lower() in ('count', 'counts', 'cnt', 'cnts'):
            if self.spec._counts is None:
                raise PlottingError('Spectrum has counts not defined')
            self._ymode = 'counts'
        elif mode.lower() == 'cps':
            self._ymode = 'cps'
        elif mode.lower() == 'cpskev':
            self._ymode = 'cpskev'
        else:
            raise PlottingError('Unknown y data mode: {}'.format(mode))

        if self._ymode == 'counts':
            self._ydata = self.spec.counts_vals
            if self._ylabel in ['Countrate [1/s]', 'Countrate [1/s/keV]'] or self._ylabel is None:
                self._ylabel = 'Counts'
        elif self._ymode == 'cps':
            self._ydata = self.spec.cps_vals
            if self._ylabel in ['Counts', 'Countrate [1/s/keV]'] or self._ylabel is None:
                self._ylabel = 'Countrate [1/s]'
        elif self._ymode == 'cpskev':
            self._ydata = self.spec.cpskev_vals
            if self._ylabel in ['Counts', 'Countrate [1/s]'] or self._ylabel is None:
                self._ylabel = 'Countrate [1/s/keV]'


    @property
    def ax(self):
        """
        Returns the current matplotlib axes object used for plotting.
        If no axes object is defined yet it will create one.
        """

        if self._ax is None:
            _, self._ax = plt.subplots()
        return self._ax


    @ax.setter
    def ax(self, ax):
        """
        Defines the current matplotlib axes object used for plotting.

        Args:
          ax: Axes to be set
        """

        self._ax = ax

    @property
    def xlabel(self):
        """
        Returns the current x label
        """
        return self._xlabel


    @xlabel.setter
    def xlabel(self, label):
        """
        Sets the xlabel to a costum value.
        """

        if label is not None:
            self._xlabel = label


    @property
    def ylabel(self):
        """
        Returns the current y label
        """

        return self._ylabel


    @ylabel.setter
    def ylabel(self, label):
        """
        Sets the ylabel to a costum value.
        """

        if label is not None:
            self._ylabel = label

    @property
    def yerror(self):
        """
        Returns array of statistical errors for current mode
        """

        if self._ymode == 'counts':
            return self.spec.counts_uncs
        elif self._ymode == 'cps':
            return self.spec.cps_uncs
        elif self._ymode == 'cpskev':
            return self.spec.cpskev_uncs


    def get_corners(self):
        """
        Creates a stepped version of the current spectrum data.

        Return:
          xcorner, ycorner: x and y values that can be used directly in
                            matplotlib's plotting function
        """

        return self.bin_edges_and_heights_to_steps(
            self._xedges, unumpy.nominal_values(self._ydata))


    def _prepare_plot(self, **kwargs):
        """Prepare for the plotting."""

        self.kwargs.update(**kwargs)
        if not self.ax.get_xlabel():
            self.ax.set_xlabel(self._xlabel)
        if not self.ax.get_ylabel():
            self.ax.set_ylabel(self._ylabel)
        if self.yscale is not None:
            self.ax.set_yscale(self.yscale)
        if self.title is not None:
            self.ax.set_title(self.title)
        elif self.spec.infilename is not None:
            self.ax.set_title(self.spec.infilename)
        if self._xlim is not None:
            self.ax.set_xlim(self.xlim)
        if self._ylim is not None:
            self.ax.set_ylim(self.ylim)
            if self.yscale == 'symlog' and self._ylim == 'default':
                self.ax.set_yscale(self.yscale, linthreshy=self.linthreshy)
        return self.get_corners()


    def plot(self, *fmt, **kwargs):
        """
        Create actual plot with matplotlib's plot method.

        Args:
          fmt:    Matplotlib plot like format string
          kwargs: Any matplotlib plot() keyword argument, overwrites
                  previously defined keywords
        """

        if hasattr(fmt, '__len__') and len(fmt) > 0:
            self.fmt = fmt

        if not hasattr(self.fmt, '__len__') or not len(self.fmt) in [0, 1]:
            raise PlottingError("Wrong number of positional argument")

        xcorners, ycorners = self._prepare_plot(**kwargs)
        self.ax.plot(xcorners, ycorners, *self.fmt, **self.kwargs)
        return self.ax


    def fill_between(self, **kwargs):
        """
        Create actual plot with matplotlib's fill_between method.

        Args:
          kwargs: Any matplotlib fill_between() keyword argument, overwrites
                  previously defined keywords
        """

        xcorners, ycorners = self._prepare_plot(**kwargs)
        self.ax.fill_between(xcorners, ycorners, **self.kwargs)
        return self.ax


    def errorbar(self, **kwargs):
        """
        Create errorbars with matplotlib's plot errorbar method.

        Args:
          kwargs: Any matplotlib errorbar() keyword argument, overwrites
                  previously defined keywords
        """

        self._prepare_plot(**kwargs)
        xdata = (self._xedges[0:-1]+self._xedges[1:])*0.5

        if 'fmt' in self.kwargs:
            self.fmt = (self.kwargs.pop('fmt'))

        if hasattr(self.fmt, '__len__') and len(self.fmt) == 0:
            self.fmt = (',',)

        if not hasattr(self.fmt, '__len__') or len(self.fmt) != 1:
            raise PlottingError("Wrong number of argument for fmt")

        self.ax.errorbar(xdata, self._ydata, yerr=self.yerror, fmt=self.fmt[0], **self.kwargs)


    def errorband(self, **kwargs):
        """
        Create an errorband with matplotlib's plot fill_between method.

        Args:
          kwargs: Any matplotlib fill_between() keyword argument, overwrites
                  previously defined keywords
        """

        self._prepare_plot(**kwargs)

        alpha = 0.5
        if 'alpha' in self.kwargs:
            alpha = self.kwargs.pop("alpha")

        xcorners, ycorlow = self.bin_edges_and_heights_to_steps(
            self._xedges, unumpy.nominal_values(self._ydata-self.yerror))
        _, ycorhig = self.bin_edges_and_heights_to_steps(
            self._xedges, unumpy.nominal_values(self._ydata+self.yerror))
        self.ax.fill_between(xcorners, ycorlow, ycorhig, alpha=alpha, **self.kwargs)
        return self.ax


    @staticmethod
    def bin_edges_and_heights_to_steps(bin_edges, heights):
        """A robust alternative to matplotlib's drawstyle='steps-*'"""

        assert len(bin_edges) == len(heights) + 1
        x = np.zeros(len(bin_edges) * 2)
        y = np.zeros_like(x)
        x[::2] = bin_edges.astype(float)
        x[1::2] = bin_edges.astype(float)
        y[1:-1:2] = heights.astype(float)
        y[2:-1:2] = heights.astype(float)
        return x, y


    @staticmethod
    def dynamic_min(data_min, min_delta_y):
        """Get an axes lower limit (for y) based on data value.

        The lower limit is the next power of 10, or 3 * power of 10, below the min.

        Args:
          data_min: the minimum of the data (could be integers or floats)
          min_delta_y: the minimum step in y
        """

        if data_min > 0:
            ceil10 = 10**(np.ceil(np.log10(data_min)))
            sig_fig = np.floor(10 * data_min / ceil10)
            if sig_fig <= 3:
                ymin = ceil10 / 10
            else:
                ymin = ceil10 / 10 * 3
        elif data_min == 0:
            ymin = min_delta_y / 10.0
        else:
            # negative
            floor10 = 10**(np.floor(np.log10(-data_min)))
            sig_fig = np.floor(-data_min / floor10)
            if sig_fig < 3:
                ymin = -floor10 * 3
            else:
                ymin = -floor10 * 10

        return ymin


    @staticmethod
    def dynamic_max(data_max, yscale):
        """Get an axes upper limit (for y) based on data value.

        The upper limit is the next power of 10, or 3 * power of 10, above
        the max (For linear, the next N * power of 10.).

        Args:
          data_max: the maximum of the data (could be integers or floats)
        """

        floor10 = 10**(np.floor(np.log10(data_max)))
        sig_fig = np.ceil(data_max / floor10)
        if yscale == 'linear':
            sig_fig = np.floor(data_max / floor10)
            ymax = floor10 * (sig_fig + 1)
        elif sig_fig < 3:
            ymax = floor10 * 3
        else:
            ymax = floor10 * 10

        return np.maximum(ymax, 0)


    @property
    def xlim(self):
        """Returns the xlim, requires xedges."""

        if self._xlim is None or self._xlim == 'default':
            return np.min(self._xedges), np.max(self._xedges)

        return self._xlim

    @xlim.setter
    def xlim(self, limits):
        """Sets xlim."""

        if limits is not None and limits != 'default' and \
        (not hasattr(limits, '__len__') or len(limits) != 2):
            raise PlottingError('xlim should be length 2: {}'.format(limits))
        self._xlim = limits


    @property
    def ylim(self):
        """Returns ylim, requires yscale, ydata."""

        if self._ylim is None or self._ylim == 'default':
            yscale = self.yscale
            if yscale is None:
                yscale = self.ax.get_yscale()

            min_ind = np.argmin(np.abs(self._ydata[self._ydata != 0]))
            delta_y = np.abs(self._ydata - self._ydata[min_ind])
            min_delta_y = np.min(delta_y[delta_y > 0])

            data_min = np.min(self._ydata)
            if yscale == 'linear':
                ymin = 0
            elif yscale == 'log' and data_min < 0:
                raise PlottingError('Cannot plot negative values on a log ' +
                                    'scale; use symlog scale')
            elif yscale == 'symlog' and data_min >= 0:
                ymin = 0
            else:
                ymin = self.dynamic_min(data_min, min_delta_y)

            data_max = np.max(self._ydata)
            ymax = self.dynamic_max(data_max, yscale)
            return ymin, ymax

        return self._ylim


    @ylim.setter
    def ylim(self, limits):
        """Sets ylim."""

        if limits is not None and limits != 'default' and \
        (not hasattr(limits, '__len__') or len(limits) != 2):
            raise PlottingError('ylim should be length 2: {}'.format(limits))
        self._ylim = limits


    @property
    def linthreshy(self):
        """Returns linthreshy, requires ydata."""

        if self._linthreshy is not None:
            return self._linthreshy
        min_ind = np.argmin(np.abs(self._ydata[self._ydata != 0]))
        delta_y = np.abs(self._ydata - self._ydata[min_ind])
        return np.min(delta_y[delta_y > 0])



"""Becquerel package metadata."""

__description__ = "Tools for radiation spectral analysis."
__url__ = "https://github.com/lbl-anp/becquerel"
__version__ = "0.2.4"
__license__ = """*** License Agreement ***

*** Copyright Notice ***

Becquerel v. 0.2.4, Copyright (c) 2017, The Regents of the University of California (UC), through Lawrence Berkeley National Laboratory, and the UC Berkeley campus (subject to receipt of any required approvals from the U.S. Dept. of Energy). All rights reserved.
If you have questions about your rights to use or distribute this software, please contact Berkeley Lab's Innovation & Partnerships Office at  IPO@lbl.gov.

NOTICE.  This Software was developed under funding from the U.S. Department of Energy and the U.S. Government consequently retains certain rights.  As such, the U.S. Government has been granted for itself and others acting on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software to reproduce, distribute copies to the public, prepare derivative works, and perform publicly and display publicly, and to permit other to do so.

**************************************************************************

 Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

 (1) Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

 (2) Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

 (3) Neither the name of the University of California, University of California, Berkeley, Lawrence Berkeley National Laboratory, or the U.S. Dept. of Energy nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

You are under no obligation whatsoever to provide any bug fixes, patches, or upgrades to the features, functionality or performance of the source code ("Enhancements") to anyone; however, if you choose to make your Enhancements available either publicly, or directly to Lawrence Berkeley National Laboratory, without imposing a separate written license agreement for such Enhancements, then you hereby grant the following license to Lawrence Berkeley National Laboratory, University if California, Berkeley, and the U.S. Dept. of Energy: a  non-exclusive, royalty-free perpetual license to install, use, modify, prepare derivative works, incorporate into other computer software, distribute, and sublicense such enhancements or derivative works thereof, in binary and source code form.

***************************************************************************
"""
__copyright__ = """

Becquerel v. 0.2.4, Copyright (c) 2017, The Regents of the University of
California (UC), through Lawrence Berkeley National Laboratory, and the UC
Berkeley campus (subject to receipt of any required approvals from the U.S.
Dept. of Energy). All rights reserved. If you have questions about your rights
to use or distribute this software, please contact Berkeley Lab's Innovation &
Partnerships Office at  IPO@lbl.gov.

NOTICE.  This Software was developed under funding from the U.S. Department of
Energy and the U.S. Government consequently retains certain rights.  As such,
the U.S. Government has been granted for itself and others acting on its
behalf a paid-up, nonexclusive, irrevocable, worldwide license in the Software
to reproduce, distribute copies to the public, prepare derivative works, and
perform publicly and display publicly, and to permit other to do so.
"""



"""Automatic calibration by associating peaks with energies."""

from __future__ import print_function
from itertools import combinations
import matplotlib.pyplot as plt
import numpy as np
from .peakfinder import PeakFinder
from .energycal import LinearEnergyCal


MAJOR_BACKGROUND_LINES = [
    1460.82,
    2614.51,
    1620.50,
    911.20,
    338.32,
    238.63,
    2204.06,
    1764.49,
    1377.67,
    1238.12,
    1120.29,
    768.36,
    609.32,
    351.93,
    295.22,
    242.00,
    186.21,
]


class AutoCalibratorError(Exception):
    """Base class for errors in AutoCalibrator."""

    pass


def fit_gain(channels, snrs, energies):
    """Calculate the mean gain to relate channels to energies.

    Assume that the peak channel error scales like fwhm/snr,
    which is proportional to sqrt(x)/snr.

    Minimize:
        F = sum(snr^2 * (x * g - y)^2 / x)
        dF/dg = sum(2 * snr^2 * x * (x * g - y) / x)
              = 2 * sum(snr^2 * (x * g - y))
              = 2 * (g * sum(snr^2 * x) - sum(snr^2 * y))
              = 0
        g = sum(snr^2 * y) / sum(snr^2 * x)

    """
    if len(channels) != len(energies):
        raise AutoCalibratorError(
            'Number of channels ({}) must equal # of energies ({})'.format(
                len(channels), len(energies)))
    if len(channels) != len(snrs):
        raise AutoCalibratorError(
            'Number of channels ({}) must equal # of SNRs ({})'.format(
                len(channels), len(snrs)))
    x = np.asarray(channels)
    s = np.asarray(snrs)
    y = np.asarray(energies)
    S2X = (s**2 * x).sum()
    S2Y = (s**2 * y).sum()
    gain = S2Y / S2X
    return gain


def fom_gain(channels, snrs, energies):
    """Calculate a leave-one-out cross-validation figure of merit.

    Assume that the peak channel error scales like fwhm/snr,
    which is proportional to sqrt(x)/snr. Then the figure of merit should be:

    F = sum(snr^2 * (x * g - y)^2 / x)

    s2x = sum(snr^2 x)
    s2y = sum(snr^2 y)

    Least-square gain when excluding channel j:
    g_j = sum_(i!=j)(snr_i^2 y_i) / sum_(i!=j)(snr_i^2 x_i)
        = (s2y - snr_j^2 y_j)/(s2x - snr_j^2 x_j)

    Figure of merit when excluding channel j:
    F_j = snr_j^2 (g_j x_j - y_j)^2 / x_j

    Leave-one-out cross validation FOM:
    F_LOO = sum_j(F_j) / N
    = sum_j(snr_j^2 (g_j x_j - y_j)^2 / x_j) / N
    = sum_j(snr_j^2 (s2y x_j - s2x y_j)^2 / (s2x - snr_j^2 x_j)^2 / x_j) / N

    """
    if len(channels) != len(energies):
        raise AutoCalibratorError(
            'Number of channels ({}) must equal # of energies ({})'.format(
                len(channels), len(energies)))
    if len(channels) != len(snrs):
        raise AutoCalibratorError(
            'Number of channels ({}) must equal # of SNRs ({})'.format(
                len(channels), len(snrs)))
    x = np.asarray(channels)
    s = np.asarray(snrs)
    y = np.asarray(energies)
    N = len(channels)
    S2X = (s**2 * x).sum()
    S2Y = (s**2 * y).sum()
    squared_errs = s**2 * (S2Y * x - S2X * y)**2 / (S2X - s**2 * x)**2 / x
    return squared_errs.sum() / N / (N - 1)


def find_best_gain(
        channels, snrs, required_energies, optional=(),
        gain_range=(1e-3, 1e3), de_max=10., verbose=False):
    """Find the gain that gives the best match of peaks to energies."""
    if len(channels) != len(snrs):
        raise AutoCalibratorError(
            'Number of channels ({}) must equal # of SNRs ({})'.format(
                len(channels), len(snrs)))
    if len(channels) < 2:
        raise AutoCalibratorError(
            'Number of channels ({}) must be at least 2'.format(len(channels)))
    if len(required_energies) < 2:
        raise AutoCalibratorError(
            'Number of required energies ({}) must be at least 2'.format(
                len(required_energies)))
    if len(channels) < len(required_energies):
        raise AutoCalibratorError(
            'Number of channels ({}) must be >= required energies ({})'.format(
                len(channels), len(required_energies)))

    channels = np.array(channels)
    snrs = np.array(snrs)
    n_req = len(required_energies)
    # make sure the required and optional sets do not overlap
    optional = sorted(list(set(optional) - set(required_energies)))
    n_opt = len(optional)
    n_set = n_req + n_opt
    best_fom = None
    best_gain = None
    best_chans = None
    best_snrs = None
    best_ergs = None
    while n_set >= n_req:
        if verbose:
            print('Searching groups of {}'.format(n_set))
        # cycle through energy combinations
        for comb_erg in combinations(optional, n_set - n_req):
            comb_erg = np.array(comb_erg)
            comb_erg = np.append(required_energies, comb_erg)
            comb_erg.sort()
            # use gain_range to reduce possible channel matches
            min_chan = (comb_erg / gain_range[1]).min()
            max_chan = (comb_erg / gain_range[0]).max()
            # cycle through channel combinations
            chan_inds = np.arange(len(channels))
            chan_inds = chan_inds[
                (min_chan <= channels) & (channels <= max_chan)]
            for chan_indices in combinations(chan_inds, n_set):
                chan_indices = np.array(chan_indices, dtype=int)
                comb_chan = np.array(channels)[chan_indices]
                comb_snr = np.array(snrs)[chan_indices]
                ind = np.argsort(comb_chan)
                comb_chan = comb_chan[ind]
                comb_snr = comb_snr[ind]
                # calculate gain
                gain = fit_gain(comb_chan, comb_snr, comb_erg)
                if gain < gain_range[0] or gain > gain_range[1]:
                    continue
                # calculate predicted energies
                pred_erg = gain * comb_chan
                de = pred_erg - comb_erg
                if (abs(de) > de_max).any():
                    continue
                # calculate figure of merit
                fom = fom_gain(comb_chan, comb_snr, comb_erg)
                if verbose:
                    print('v')
                    s0 = 'Valid calibration found:\n'
                    s0 += 'FOM: {:15.9f}'.format(fom)
                    s0 += '  gain: {:6.3f}'.format(gain)
                    s0 += '  ergs: {:50s}'.format(str(comb_erg))
                    s0 += '  de: {:50s}'.format(str(de))
                    s0 += '  chans: {:40s}'.format(str(comb_chan))
                    print(s0)
                if best_fom is None:
                    best_fom = fom + 1.
                if fom < best_fom:
                    best_fom = fom
                    best_gain = gain
                    best_chans = comb_chan
                    best_snrs = comb_snr
                    best_ergs = comb_erg
                    if verbose:
                        s0 = 'Best calibration so far:\n'
                        s0 += 'FOM: {:15.9f}'.format(best_fom)
                        s0 += '  gain: {:6.3f}'.format(best_gain)
                        s0 += '  ergs: {:50s}'.format(str(best_ergs))
                        s0 += '  de: {:50s}'.format(str(de))
                        s0 += '  chans: {:40s}'.format(str(best_chans))
                        print(s0)
        n_set -= 1
    if best_gain is None:
        return None
    else:
        print('found best gain: %f keV/channel' % best_gain)
        return {
            'gain': best_gain,
            'channels': best_chans,
            'snrs': best_snrs,
            'energies': best_ergs,
        }


class AutoCalibrator(object):
    """Automatically calibrate a spectrum by convolving it with a filter.

    A note on nomenclature: for historic reasons, 'channels' is used in
    autocal.py for generic uncalibrated x-axis values. A 'channel' is no longer
    necessarily an integer channel number (i.e., bin) from a multi-channel
    analyzer, but could for instance be a float-type fC of charge collected.
    """

    def __init__(self, peakfinder):
        """Initialize the calibration with a spectrum and kernel."""
        self.set_peaks(peakfinder)
        self.gain = None
        self.cal = None
        self.fit_channels = []
        self.fit_snrs = []
        self.fit_energies = []
        self.reset()

    def reset(self):
        """Reset all of the members."""
        self.gain = None
        self.cal = None
        self.fit_channels = []
        self.fit_snrs = []
        self.fit_energies = []

    def set_peaks(self, peakfinder):
        """Use the peaks found by the PeakFinder."""
        if not isinstance(peakfinder, PeakFinder):
            raise AutoCalibratorError(
                'Argument must be a PeakFinder, not {}'.format(
                    type(peakfinder)))
        self.peakfinder = peakfinder

    def plot(self, **kwargs):
        """Plot the peaks found and the peaks used to fit."""
        self.peakfinder.plot(peaks=True, **kwargs)
        for chan, snr in zip(self.fit_channels, self.fit_snrs):
            plt.plot([chan] * 2, [0, snr], 'g-', lw=2)
            plt.plot(chan, snr, 'go')

    def fit(self, required_energies, optional=(),
            gain_range=(1e-4, 1e3), de_max=10., verbose=False):
        """Find the gain that gives the best match of peaks to energies."""
        if len(self.peakfinder.centroids) == 1 and \
                len(required_energies) == 1:
            # special case: only one line identified
            self.fit_channels = list(self.peakfinder.centroids)
            self.fit_snrs = list(self.peakfinder.snrs)
            self.fit_energies = list(required_energies)
            gain = required_energies[0] / self.peakfinder.centroids[0]
            self.gain = gain
            self.cal = LinearEnergyCal.from_coeffs(
                {'offset': 0, 'slope': self.gain})
            return
        # handle the usual case: multiple lines to match
        if len(self.peakfinder.centroids) < 2:
            raise AutoCalibratorError(
                'Need more than {} peaks to fit'.format(
                    len(self.peakfinder.centroids)))
        if len(required_energies) < 2:
            raise AutoCalibratorError(
                'Need more than {} energies to fit'.format(
                    len(required_energies)))
        if len(self.peakfinder.centroids) < len(required_energies):
            raise AutoCalibratorError(
                'Require {} energies but only {} peaks are available'.format(
                    len(required_energies), len(self.peakfinder.centroids)))
        fit = find_best_gain(
            self.peakfinder.centroids, self.peakfinder.snrs,
            required_energies, optional=optional, gain_range=gain_range,
            de_max=de_max, verbose=verbose)
        if fit is None:
            self.fit_channels = []
            self.fit_snrs = []
            self.fit_energies = []
            self.gain = None
            self.cal = None
            raise AutoCalibratorError('No valid fit was found')
        else:
            self.fit_channels = fit['channels']
            self.fit_snrs = fit['snrs']
            self.fit_energies = fit['energies']
            self.gain = fit['gain']
            self.cal = LinearEnergyCal.from_coeffs(
                {'offset': 0, 'slope': self.gain})



"""Query photon cross section data from NIST XCOM database.

Provides the fetch_xcom_data function to query the NIST XCOM database,
as well as a few mixtures for common materials.

References:
  https://www.nist.gov/pml/xcom-photon-cross-sections-database
  http://physics.nist.gov/PhysRefData/Xcom/html/xcom1.html

"""

from __future__ import print_function
import requests
import pandas as pd
from . import element
from ..core.utils import isstring, Iterable

# Dry air relative weights taken from:
# http://www.engineeringtoolbox.com/air-composition-d_212.html
MIXTURE_AIR_DRY = [
    'N2 75.47',
    'O2 23.20',
    'Ar 0.933',
    'CO2 0.03',
]

# Seawater composition for 3.5% salinity taken from:
# https://en.wikipedia.org/wiki/Seawater#Compositional_differences_from_freshwater
MIXTURE_SEAWATER = [
    'O 85.84',
    'H 10.82',
    'Cl 1.94',
    'Na 1.08',
    'Mg 0.1292',
    'S 0.091',
    'Ca 0.04',
    'K 0.04',
    'Br 0.0067',
    'C 0.0028',
]

# typical Portland cement values taken from:
# http://matse1.matse.illinois.edu/concrete/prin.html
MIXTURE_PORTLAND_CEMENT = [
    'Ca3SiO5 50',
    'Ca2SiO4 25',
    'Ca3Al2O6 10',
    'Ca4Al2Fe2O10 10',
    'CaSO6H4 5',
]


# base URL for three scripts on NIST where data can be queried
_URL = 'https://physics.nist.gov/cgi-bin/Xcom/xcom3_'


# Dictionary of data that will be posted to URL
_DATA = {
    'NumAdd': '1',          # always '1'?
    'Energies': '',         # additional energies separated by ; (MeV)
    'WindowXmin': '0.001',  # lower limit of energy grid (MeV)
    'WindowXmax': '100',    # upper limit of energy grid (MeV)
    'Output': '',           # 'on' for standard energy grid
    'OutOpt': 'PIC',        # return cross sections in cm^2/g
    'ResizeFlag': 'on',     # seems to determine whether Xmin and Xmax are used
}

# abbreviated names for the table columns (these are used in the dataframe)
COLUMNS_SHORT = ['energy', 'coherent', 'incoherent', 'photoelec',
                 'pair_nuc', 'pair_elec', 'total_w_coh', 'total_wo_coh']

# medium length names for the table columns
COLUMNS_MEDIUM = {
    'energy': 'Energy',
    'coherent': 'Coherent',
    'incoherent': 'Incoherent',
    'photoelec': 'Photoelectric',
    'pair_nuc': 'Pair production (nuclear)',
    'pair_elec': 'Pair Production (electron)',
    'total_w_coh': 'Total with coherent',
    'total_wo_coh': 'Total without coherent'
}

# verbose names for the table columns
COLUMNS_LONG = {
    'energy': 'Photon Energy',
    'coherent': 'Coherent Scattering',
    'incoherent': 'Incoherent Scattering',
    'photoelec': 'Photoelectric Absorption',
    'pair_nuc': 'Pair Production in Nuclear Field',
    'pair_elec': 'Pair Production in Electron Field',
    'total_w_coh': 'Total Attenuation with Coherent Scattering',
    'total_wo_coh': 'Total Attenuation without Coherent Scattering'
}


class XCOMError(Exception):
    """General XCOM error."""

    pass


class XCOMInputError(XCOMError):
    """Error related to the user input to XCOMQuery."""

    pass


class XCOMRequestError(XCOMError):
    """Error related to communicating with XCOM or parsing the result."""

    pass


class _XCOMQuery(object):
    """Query photon cross section data from NIST XCOM database.

    After the data have been successfully queried, they are stored in a
    pandas DataFrame that is accessible through this class's methods.

    The DataFrame column names amd units are:
      'energy': Photon Energy [keV]
      'coherent': coherent scattering cross section [cm^2/g]
      'incoherent': incoherent scattering cross section [cm^2/g]
      'photoelec': photoelectric absorption cross section [cm^2/g]
      'pair_nuc': pair production cross section in nuclear field [cm^2/g]
      'pair_elec': pair production cross section in electron field [cm^2/g]
      'total_w_coh': total attenuation with coherent scattering [cm^2/g]
      'total_wo_coh': total attenuation without coherent scattering [cm^2/g]

    Methods:
      update: add or change the search criteria
      perform: perform the query and parse the results
      __getitem__: use [] to get the DataFrame column
      __len__: len() returns the length of the DataFrame
      __str__: str() returns the str method of the DataFrame
      __format__: format() uses the format method of the DataFrame
      keys: returns the DataFrame keys

    """

    def __init__(self, arg, **kwargs):
        """Initialize and perform an XCOM query.

        Args:
          arg: the atomic number, element symbol, compound string, or mixture
            (the type of argument will be inferred from its content).
          e_range_kev (optional): a length-2 iterable giving the lower and
            upper bounds for a standard grid of energies in keV. Limits must be
            between 1 keV and 1E8 keV, inclusive.
          energies_kev (optional): an iterable of specific energies in keV at
            which cross sections will be evaluated. Energies must be between
            1 keV and 1E8 keV, inclusive.
          perform (optional): set to False to prevent query from immediately
            being performed. [default: True]

        Raises:
          XCOMInputError: if bad search criteria are set or search criteria
            are incomplete.
          XCOMRequestError: if there is a problem with the URL request, or
            a problem parsing the data.

        """
        self._url = _URL
        self._req = None
        self._text = None
        self.df = None
        self._data = dict(_DATA)
        self._method = ''
        # determine which kind of argument 'arg' is (symbol, Z, compound, mix)
        kwargs.update(_XCOMQuery._argument_type(arg))
        self.update(**kwargs)
        if kwargs.get('perform', True):
            self.perform()

    def __len__(self):
        """Pass-through to use DataFrame len()."""
        if self.df is None:
            return 0
        elif len(self.df.keys()) == 0:
            return 0
        else:
            return len(self.df[self.df.keys()[0]])

    def keys(self):
        """Pass-through for DataFrame keys method."""
        return self.df.keys()

    def __getitem__(self, key):
        """Pass-through so that [] accesses the DataFrame."""
        return self.df[key]

    def __str__(self):
        """Pass-through to use DataFrame str method."""
        return str(self.df)

    def __format__(self, formatstr):
        """Pass-through to use DataFrame format method."""
        return self.df.__format__(formatstr)

    @staticmethod
    def _argument_type(arg):
        """Determine if argument is a symbol, Z, compound, or mixture."""
        if isstring(arg):
            if arg.isdigit():
                return {'z': arg}
            elif arg.lower() in [s.lower() for s in element.SYMBOLS]:
                return {'symbol': arg}
            else:
                return {'compound': arg}
        elif isinstance(arg, int):
            return {'z': arg}
        elif isinstance(arg, Iterable):
            return {'mixture': arg}
        raise XCOMInputError(
            'Cannot determine if argument {}'.format(arg) +
            ' is a symbol, Z, compound, or mixture')

    @staticmethod
    def _check_z(zstr):
        """Check whether the Z is valid. Raise XCOMInputError if not."""
        zint = int(zstr)
        if zint < 1 or zint > 100:
            raise XCOMInputError(
                'XCOM only supports Z from 1 to 100 (z={})'.format(zstr))

    @staticmethod
    def _check_compound(formula):
        """Check whether the compound is valid. Raise XCOMInputError if not."""
        if not formula.isalnum():
            raise XCOMInputError('Formula not valid: {}'.format(formula))

    @staticmethod
    def _check_mixture(formulae):
        """Check whether the mixture is valid. Raise XCOMInputError if not."""
        if not isinstance(formulae, Iterable):
            raise XCOMInputError(
                'Mixture formulae must be an iterable: {}'.format(formulae))
        for formula in formulae:
            try:
                compound, weight = formula.split()
            except AttributeError:
                raise XCOMInputError(
                    'Mixture formulae "{}" line "{}" must be a string'.format(
                        formulae, formula))
            except ValueError:
                raise XCOMInputError(
                    'Mixture formulae "{}" line "{}" must split into 2'.format(
                        formulae, formula))
            _XCOMQuery._check_compound(compound)
            try:
                float(weight)
            except (ValueError, TypeError):
                raise XCOMInputError(
                    'Mixture formulae "{}" has bad weight "{}"'.format(
                        formulae, weight))

    def update(self, **kwargs):
        """Update the search criteria.

        Before calling perform(), one of the following search criteria must
        be set here: z, symbol, compound, or mixture. A valid query will also
        require setting either e_range_kev and/or energies_kev.

        Args:
          symbol (optional): a string of the element symbol, e.g., 'Ge'.
          z (optional): an integer of the element atomic number, e.g., 32.
          compound (optional): a string of the chemical formula, e.g., 'H2O'.
          mixture (optional): a list of compounds and relative weights, e.g.,
            ['H2O 0.5', 'Ge 0.5']
          e_range_kev (optional): a length-2 iterable giving the lower and
            upper bounds for a standard grid of energies in keV. Limits must be
            between 1 keV and 1E8 keV, inclusive.
          energies_kev (optional): an iterable of specific energies in keV at
            which cross sections will be evaluated. Energies must be between
            1 keV and 1E8 keV, inclusive.

        Raises:
          XCOMInputError: if bad search criteria are set.

        """
        # check for valid keywords
        for kwarg in kwargs:
            if kwarg not in [
                    'symbol', 'z', 'compound', 'mixture',
                    'e_range_kev', 'energies_kev', 'perform']:
                raise XCOMInputError('Unknown keyword: "{}"'.format(kwarg))

        # determine the search method (element, compound, or mixture)
        if 'symbol' in kwargs:
            self._method = '1'
            sym = kwargs['symbol']
            self._data['ZSym'] = sym
        elif 'z' in kwargs:
            self._method = '1'
            znum = kwargs['z']
            _XCOMQuery._check_z(znum)
            self._data['ZNum'] = '{:d}'.format(int(znum))
        elif 'compound' in kwargs:
            # convert compound to mixture to avoid occasional problem
            # with XCOM compound queries (see issue #76)
            # self._method = '2'
            self._method = '3'
            formula = kwargs['compound']
            _XCOMQuery._check_compound(formula)
            formulae = [formula + ' 1']
            _XCOMQuery._check_mixture(formulae)
            formulae = '\r\n'.join(formulae)
            # self._data['Formula'] = formula
            self._data['Formulae'] = formulae
        elif 'mixture' in kwargs:
            self._method = '3'
            formulae = kwargs['mixture']
            _XCOMQuery._check_mixture(formulae)
            formulae = '\r\n'.join(formulae)
            self._data['Formulae'] = formulae

        # include standard grid of energies
        if 'e_range_kev' in kwargs:
            if not isinstance(kwargs['e_range_kev'], Iterable):
                raise XCOMInputError(
                    'XCOM e_range_kev must be iterable of length 2: {}'.format(
                        kwargs['e_range_kev']))
            if len(kwargs['e_range_kev']) != 2:
                raise XCOMInputError(
                    'XCOM e_range_kev must be iterable of length 2: {}'.format(
                        kwargs['e_range_kev']))
            if kwargs['e_range_kev'][0] < 1:
                raise XCOMInputError(
                    'XCOM e_range_kev[0] must be >= 1 keV: {}'.format(
                        kwargs['e_range_kev'][0]))
            if kwargs['e_range_kev'][1] > 1e8:
                raise XCOMInputError(
                    'XCOM e_range_kev[1] must be <= 1E8 keV: {}'.format(
                        kwargs['e_range_kev'][1]))
            if kwargs['e_range_kev'][0] >= kwargs['e_range_kev'][1]:
                raise XCOMInputError(
                    'XCOM e_range_kev[0] must be < e_range_kev[1]: {}'.format(
                        kwargs['e_range_kev']))
            self._data['WindowXmin'] = '{:.6f}'.format(
                kwargs['e_range_kev'][0] / 1000.)
            self._data['WindowXmax'] = '{:.6f}'.format(
                kwargs['e_range_kev'][1] / 1000.)
            self._data['Output'] = 'on'

        # additional energies
        if 'energies_kev' in kwargs:
            if not isinstance(kwargs['energies_kev'], Iterable):
                raise XCOMInputError(
                    'XCOM energies_kev must be an iterable: {}'.format(
                        kwargs['energies_kev']))
            for energy in kwargs['energies_kev']:
                if energy < 1 or energy > 1e8:
                    raise XCOMInputError(
                        'XCOM energy must be >= 1 and <= 1E8 keV: {}'.format(
                            energy))
            self._data['Energies'] = ';'.join([
                '{:.6f}'.format(erg / 1000.)
                for erg in kwargs['energies_kev']])

    def _request(self):
        """Request data table from the URL."""
        self._req = requests.post(self._url + self._method, data=self._data)
        if not self._req.ok or self._req.reason != 'OK' or \
                self._req.status_code != 200:
            raise XCOMRequestError(
                'XCOM Request failed: reason={}, status_code={}'.format(
                    self._req.reason, self._req.status_code))
        if 'Error' in self._req.text:
            raise XCOMRequestError(
                'XCOM returned an error:\n{}'.format(self._req.text))

    def _parse_text(self):
        """Parse table contained in the text into a dictionary."""
        self._text = str(self._req.text)
        if len(self._text) == 0:
            raise XCOMRequestError('XCOM returned no text')
        tables = pd.read_html(self._text, header=0, skiprows=[1, 2])
        if len(tables) != 1:
            raise XCOMRequestError('More than one HTML table found')
        self.df = tables[0]
        if len(self.df.keys()) != 1 + len(COLUMNS_SHORT):
            raise XCOMRequestError(
                'Found {} columns but expected {}'.format(
                    len(self.df.keys()), 1 + len(COLUMNS_SHORT)))
        # remove 'edge' column
        self.df = self.df[self.df.keys()[1:]]
        self.df.columns = COLUMNS_SHORT
        if len(self) == 0:
            raise XCOMRequestError('Parsed DataFrame is empty')

    def perform(self):
        """Perform the query.

        Before calling perform(), set the search criteria using update()
        or __init__().

        Raises:
          XCOMInputError: if search criteria are incomplete.
          XCOMRequestError: if there is a problem with the URL request, or
            a problem parsing the data.

        """
        if self._method not in ['1', '2', '3']:
            raise XCOMInputError(
                'XCOM search method not set. Need to call update() method.')
        if self._data['Energies'] == '' and self._data['Output'] == '':
            raise XCOMInputError('No energies_kev or e_range_kev requested.')
        # submit the query
        self._request()
        # package the output into a pandas DataFrame
        self._parse_text()
        # convert energy from MeV to keV
        self.df['energy'] *= 1000.


def fetch_xcom_data(arg, **kwargs):
    """Query photon cross section data from NIST XCOM database.

    Returns a pandas DataFrame containing the cross section data.

    The DataFrame column names amd units are:
      'energy': Photon Energy [keV]
      'coherent': coherent scattering cross section [cm^2/g]
      'incoherent': incoherent scattering cross section [cm^2/g]
      'photoelec': photoelectric absorption cross section [cm^2/g]
      'pair_nuc': pair production cross section in nuclear field [cm^2/g]
      'pair_elec': pair production cross section in electron field [cm^2/g]
      'total_w_coh': total attenuation with coherent scattering [cm^2/g]
      'total_wo_coh': total attenuation without coherent scattering [cm^2/g]

    Args:
      arg: the atomic number, element symbol, compound string, or mixture
        (the type of argument will be inferred from its content).
      e_range_kev (optional): a length-2 iterable giving the lower and
        upper bounds for a standard grid of energies in keV. Limits must be
        between 1 keV and 1E8 keV, inclusive.
      energies_kev (optional): an iterable of specific energies in keV at
        which cross sections will be evaluated. Energies must be between
        1 keV and 1E8 keV, inclusive.

    Raises:
      XCOMInputError: if bad search criteria are set or search criteria
        are incomplete.
      XCOMRequestError: if there is a problem with the URL request, or
        a problem parsing the data.

    """
    query = _XCOMQuery(arg, **kwargs)
    return query.df



"""Tools for radiation spectral analysis."""

from . import xcom, nndc, materials, element, isotope, isotope_qty
from .element import (Element, ElementError, ElementZError,
                      ElementSymbolError, ElementNameError)
from .isotope import Isotope, IsotopeError
from .isotope_qty import (IsotopeQuantity, IsotopeQuantityError,
                          NeutronIrradiation, NeutronIrradiationError,
                          UCI_TO_BQ, N_AV)
from .materials import (fetch_element_data, fetch_compound_data,
                        NISTMaterialsError, NISTMaterialsRequestError)
from .nndc import (fetch_wallet_card, fetch_decay_radiation,
                   NNDCError, NoDataFound, NNDCInputError, NNDCRequestError)
from .xcom import (fetch_xcom_data,
                   XCOMError, XCOMInputError, XCOMRequestError,
                   MIXTURE_AIR_DRY, MIXTURE_SEAWATER, MIXTURE_PORTLAND_CEMENT)

__all__ = [
    'xcom', 'nndc', 'materials', 'element', 'isotope', 'isotope_qty',
    'Element', 'ElementError', 'ElementZError',
    'ElementSymbolError', 'ElementNameError',
    'Isotope', 'IsotopeError',
    'IsotopeQuantity', 'IsotopeQuantityError',
    'NeutronIrradiation', 'NeutronIrradiationError',
    'UCI_TO_BQ', 'N_AV',
    'fetch_element_data', 'fetch_compound_data',
    'NISTMaterialsError', 'NISTMaterialsRequestError',
    'fetch_wallet_card', 'fetch_decay_radiation',
    'NNDCError', 'NoDataFound', 'NNDCInputError', 'NNDCRequestError',
    'fetch_xcom_data',
    'XCOMError', 'XCOMInputError', 'XCOMRequestError',
    'MIXTURE_AIR_DRY', 'MIXTURE_SEAWATER', 'MIXTURE_PORTLAND_CEMENT',
]



"""A cache of all NNDC wallet card data."""

from __future__ import print_function
from future.builtins import super
import pandas as pd
import uncertainties
from . import nndc
from . import df_cache
from ..core.utils import isstring

# pylint: disable=no-self-use


def convert_float_ufloat(x):
    """Convert string to a float or a ufloat, including None ('') and NaN.

    Args:
      x: a string giving the number
    """

    if isstring(x):
        if '+/-' in x:
            tokens = x.split('+/-')
            return uncertainties.ufloat(float(tokens[0]), float(tokens[1]))
        if x == '':
            return None
    return float(x)


def format_ufloat(x, fmt='{:.12f}'):
    """Convert ufloat to a string, including None ('') and NaN.

    Args:
      x: a ufloat
    """

    if x is None:
        return ''
    return fmt.format(x)


class WalletCardCache(df_cache.DataFrameCache):
    """A cache of all isotope wallet cards from NNDC."""

    name = 'all_wallet_cards'

    def write_file(self):
        """Format ufloat columns before writing so they keep precision."""

        for col in ['Abundance (%)', 'Mass Excess (MeV)']:
            self.df[col] = self.df[col].apply(format_ufloat)
        super().write_file()

    def read_file(self):
        """Ensure some columns are properly converted to float/ufloat."""

        super().read_file()
        for col in ['Abundance (%)', 'Mass Excess (MeV)']:
            self.df[col] = self.df[col].apply(convert_float_ufloat)

    def fetch(self):
        """Fetch wallet card data from NNDC for all isotopes."""

        self.df = pd.DataFrame()
        z_edges = (0, 40, 80, 120)
        for z0, z1 in zip(z_edges[:-1], z_edges[1:]):
            df_chunk = nndc.fetch_wallet_card(z_range=(z0, z1 - 1))
            self.df = self.df.append(df_chunk)
        self.loaded = True


wallet_cache = WalletCardCache()



"""Read in an Ortec SPE file."""

from __future__ import print_function
import datetime
import os
import warnings
import dateutil.parser
import numpy as np
from .spectrum_file import (SpectrumFile, SpectrumFileParsingError,
                            SpectrumFileParsingWarning)
warnings.simplefilter('always', DeprecationWarning)


class SpeFileParsingError(SpectrumFileParsingError):
    """Failed while parsing an SPE file."""

    pass


class SpeFileWritingError(SpectrumFileParsingError):
    """Failed while writing an SPE file."""

    pass


class SpeFile(SpectrumFile):
    """SPE ASCII file parser.

    Just instantiate a class with a filename:
        spec = SpeFile(filename)

    Then the data are in
        spec.data [counts]
        spec.channels
        spec.energies
        spec.bin_edges_kev
        spec.energy_bin_widths
        spec.energy_bin_edges (deprecated)

    ORTEC's SPE file format is given on page 73 of this document:
        http://www.ortec-online.com/download/ortec-software-file-structure-manual.pdf

    """

    def __init__(self, filename):
        """Initialize the SPE file."""
        super(SpeFile, self).__init__(filename)
        _, ext = os.path.splitext(self.filename)
        if ext.lower() != '.spe':
            raise SpeFileParsingError('File extension is incorrect: ' + ext)
        # SPE-specific members
        self.first_channel = 0
        self.ROIs = []
        self.energy_cal = []
        self.shape_cal = []
        # read in the data
        self.read()
        self.apply_calibration()

    def read(self, verbose=False):
        """Read in the file."""
        print('SpeFile: Reading file ' + self.filename)
        self.realtime = 0.0
        self.livetime = 0.0
        self.channels = np.array([], dtype=np.float)
        self.data = np.array([], dtype=np.float)
        self.cal_coeff = []
        with open(self.filename, 'r') as f:
            # read & remove newlines from end of each line
            lines = [line.strip() for line in f.readlines()]
            i = 0
            while i < len(lines):
                # check whether we have reached a keyword and parse accordingly
                if lines[i] == '$SPEC_ID:':
                    i += 1
                    self.spectrum_id = lines[i]
                    if verbose:
                        print(self.spectrum_id)
                elif lines[i] == '$SPEC_REM:':
                    self.sample_description = ''
                    i += 1
                    while i < len(lines) and not lines[i].startswith('$'):
                        self.sample_description += lines[i] + '\n'
                        i += 1
                    self.sample_description = self.sample_description[:-1]
                    i -= 1
                    if verbose:
                        print(self.sample_description)
                elif lines[i] == '$DATE_MEA:':
                    i += 1
                    self.collection_start = dateutil.parser.parse(lines[i])
                    if verbose:
                        print(self.collection_start)
                elif lines[i] == '$MEAS_TIM:':
                    i += 1
                    self.livetime = float(lines[i].split(' ')[0])
                    self.realtime = float(lines[i].split(' ')[1])
                    if verbose:
                        print(self.livetime, self.realtime)
                elif lines[i] == '$DATA:':
                    i += 1
                    self.first_channel = int(lines[i].split(' ')[0])
                    # I don't know why it would be nonzero
                    if self.first_channel != 0:
                        raise SpeFileParsingError(
                            'First channel is not 0: {}'.format(
                                self.first_channel))
                    self.num_channels = int(lines[i].split(' ')[1])
                    if verbose:
                        print(self.first_channel, self.num_channels)
                    j = self.first_channel
                    while j <= self.num_channels + self.first_channel:
                        i += 1
                        self.data = np.append(self.data, int(lines[i]))
                        self.channels = np.append(self.channels, j)
                        j += 1
                elif lines[i] == '$ROI:':
                    self.ROIs = []
                    i += 1
                    while i < len(lines) and not lines[i].startswith('$'):
                        self.ROIs.append(lines[i])
                        i += 1
                    i -= 1
                    if verbose:
                        print(self.ROIs)
                elif lines[i] == '$ENER_FIT:':
                    i += 1
                    self.energy_cal.append(float(lines[i].split(' ')[0]))
                    self.energy_cal.append(float(lines[i].split(' ')[1]))
                    if verbose:
                        print(self.energy_cal)
                elif lines[i] == '$MCA_CAL:':
                    i += 1
                    n_coeff = int(lines[i])
                    i += 1
                    for j in range(n_coeff):
                        self.cal_coeff.append(float(lines[i].split(' ')[j]))
                    if verbose:
                        print(self.cal_coeff)
                elif lines[i] == '$SHAPE_CAL:':
                    i += 1
                    n_coeff = int(lines[i])
                    i += 1
                    for j in range(n_coeff):
                        self.shape_cal.append(float(lines[i].split(' ')[j]))
                    if verbose:
                        print(self.shape_cal)
                elif lines[i].startswith('$'):
                    key = lines[i][1:].rstrip(':')
                    i += 1
                    values = []
                    while i < len(lines) and not lines[i].startswith('$'):
                        values.append(lines[i])
                        i += 1
                    if i < len(lines):
                        if lines[i].startswith('$'):
                            i -= 1
                    self.metadata[key] = values
                else:
                    warnings.warn(
                        'Line {} unknown: '.format(i + 1) + lines[i],
                        SpectrumFileParsingWarning)
                i += 1
        if self.realtime <= 0.0:
            raise SpeFileParsingError(
                'Realtime not parsed correctly: {}'.format(self.realtime))
        if self.livetime <= 0.0:
            raise SpeFileParsingError(
                'Livetime not parsed correctly: {}'.format(self.livetime))
        if self.livetime > self.realtime:
            raise SpeFileParsingError(
                'Livetime > realtime: {} > {}'.format(
                    self.livetime, self.realtime))
        self.collection_stop = self.collection_start + \
            datetime.timedelta(seconds=self.realtime)

    def _spe_format(self):
        """Format of this spectrum for writing to file."""
        s = ''
        s += '$SPEC_ID:\n'
        s += self.spectrum_id + '\n'
        s += '$SPEC_REM:\n'
        s += self.sample_description + '\n'
        if self.collection_start is not None:
            s += '$DATE_MEA:\n'
            s += '{:%m/%d/%Y %H:%M:%S}\n'.format(self.collection_start)
        s += '$MEAS_TIM:\n'
        s += '{:.0f} {:.0f}\n'.format(self.livetime, self.realtime)
        s += '$DATA:\n'
        s += '{:.0f} {:d}\n'.format(self.first_channel, self.num_channels)
        for j in range(self.num_channels):
            s += '       {:.0f}\n'.format(self.data[j])
        s += '$ROI:\n'
        for line in self.ROIs:
            s += line + '\n'
        if len(self.energy_cal) > 0:
            s += '$ENER_FIT:\n'
            s += '{:f} {:f}\n'.format(self.energy_cal[0], self.energy_cal[1])
        if len(self.cal_coeff) > 0:
            s += '$MCA_CAL:\n'
            n_coeff = len(self.cal_coeff)
            s += '{:d}\n'.format(n_coeff)
            s += '{:E}'.format(self.cal_coeff[0])
            for j in range(1, n_coeff):
                s += ' {:E}'.format(self.cal_coeff[j])
            s += '\n'
        if len(self.shape_cal) > 0:
            s += '$SHAPE_CAL:\n'
            n_coeff = len(self.shape_cal)
            s += '{:d}\n'.format(n_coeff)
            s += '{:E}'.format(self.shape_cal[0])
            for j in range(1, n_coeff):
                s += ' {:E}'.format(self.shape_cal[j])
            s += '\n'
        if len(self.metadata.keys()) > 0:
            for key, values in self.metadata.items():
                s += '$' + key + ':\n'
                for val in values:
                    s += str(val) + '\n'
        return s[:-1]

    def write(self, filename):
        """Write back to a file."""
        _, ext = os.path.splitext(filename)
        if ext.lower() != '.spe':
            raise SpeFileWritingError('File extension is incorrect: ' + ext)
        with open(filename, 'w') as outfile:
            print(self._spe_format(), file=outfile)



"""Spectrum file parser classes.

Just instantiate a class with a filename:
    spec = SpeFile(filename)

Then the data are in
    spec.data [counts]
    spec.channels
    spec.energies
    spec.energy_bin_widths

"""

from .spe_file import SpeFile
from .spc_file import SpcFile
from .cnf_file import CnfFile
from .spectrum_file import SpectrumFileParsingError, SpectrumFileParsingWarning



"""Quantities of a nuclear isotope, with decay and activation tools."""

from __future__ import print_function
import datetime
import copy
import numpy as np
from .isotope import Isotope
from ..core import utils

UCI_TO_BQ = 3.7e4
N_AV = 6.022141e23


class IsotopeQuantityError(Exception):
    """Raised by the IsotopeQuantity class"""

    pass


def handle_isotope(isotope, error_name=None):
    """Handle string or Isotope input.

    Args:
      isotope: either a string of an isotope name, or an Isotope object

    Raises:
      TypeError: if isotope is not a string or Isotope
      IsotopeError: if string is bad

    Returns:
      an Isotope object
    """

    if isinstance(isotope, Isotope):
        return isotope
    elif utils.isstring(isotope):
        return Isotope(isotope)
    else:
        raise TypeError(
            '{} needs an Isotope instance or string, not {}'.format(
                error_name, isotope))


class IsotopeQuantity(object):
    """An amount of an isotope.

    Can be multiplied or divided by a scalar, to produce a copy with the same
    isotope and reference date but a scaled reference quantity.

    Two IsotopeQuantity instances are equal iff they are the same isotope
    and the quantities are np.isclose for any given datetime.

    Construction class methods:
      from_decays: activity based on number of decays in a given time interval
      from_comparison: activity by comparing to a measured known sample

    Data Attributes:
      isotope: an Isotope object, the isotope of this material
      half_life: the half life of the isotope, in seconds
      decay_const: the decay constant of the isotope, in 1/seconds
      is_stable: bool representing whether the isotope is considered stable
      ref_date: a datetime object representing the reference date/time
      ref_atoms: the number of atoms of the isotope, at the reference time.

    Methods:
      atoms_at: number of atoms at given time
      bq_at: activity in Bq at given time
      uci_at: activity in uCi at given time
      g_at: mass in grams at given time
      atoms_now, bq_now, uci_now, g_now: quantity at current time
      decays_from: number of decays during a time interval
      bq_from, uci_from: average activity during a time interval
      decays_during: number of decays during a Spectrum measurement
      bq_during, uci_during: average activity during a Spectrum measurement
      time_when: time at which activity or mass equals a given value
    """

    def __init__(self, isotope, date=None, stability=1e18, **kwargs):
        """Initialize.

        Specify one of bq, uci, atoms, g to define the quantity.

        Args:
          isotope: an Isotope object, of which this is a quantity,
            OR a string to instantiate the Isotope
          date: the reference date for the activity or mass
          stability: half-life above which an isotope is considered stable [s]
          bq: the activity at the reference date [Bq]
          uci: the activity at the reference date [uCi]
          atoms: the number of atoms at the reference date
          g: the mass at the reference date [g]

        Raises:
          TypeError: if isotope is not an Isotope object
          AttributeError: if isotope is missing half_life or decay_const
          IsotopeQuantityError: if no valid quantity kwarg specified
        """

        self._init_isotope(isotope, stability)
        self._init_date(date)
        self.ref_atoms = self._atoms_from_kwargs(**kwargs)

    def _init_isotope(self, isotope, stability):
        """Initialize the isotope.

        Args:
          isotope: an Isotope object, or a string that defines an Isotope
          stability: the half-life above which an isotope is considered
            stable [s]

        Raises:
          TypeError: if isotope is not an Isotope object
          AttributeError: if isotope is missing half_life or decay_const
        """

        self.isotope = handle_isotope(isotope, error_name='IsotopeQuantity')

        self.half_life = self.isotope.half_life
        self.decay_const = self.isotope.decay_const
        self.is_stable = self.half_life > stability

    def _init_date(self, date):
        """Initialize the reference date/time.

        Args:
          date: a date string or datetime.datetime object
        """

        self.ref_date = utils.handle_datetime(
            date, error_name='IsotopeQuantity date', allow_none=True)
        if self.ref_date is None:
            # assume a long-lived source in the current epoch
            self.ref_date = datetime.datetime.now()

    def _atoms_from_kwargs(self, **kwargs):
        """Parse kwargs and return a quantity in atoms.

        Args (specify one):
          atoms: the number of atoms
          bq: the activity [Bq]
          uci: the activity [uCi]
          g: the mass [g]
          _init_empty: (internal use only) set True if the reference quantity
            will be set later

        Raises:
          IsotopeQuantityError: if no valid argument specified
        """

        if 'atoms' in kwargs:
            return self._check_positive_qty(kwargs['atoms'])
        elif 'g' in kwargs:
            return (self._check_positive_qty(kwargs['g']) /
                    self.isotope.A * N_AV)
        elif 'bq' in kwargs and not self.is_stable:
            return (self._check_positive_qty(kwargs['bq']) /
                    self.decay_const)
        elif 'uci' in kwargs and not self.is_stable:
            return (self._check_positive_qty(kwargs['uci']) *
                    UCI_TO_BQ / self.decay_const)
        elif 'bq' in kwargs or 'uci' in kwargs:
            raise IsotopeQuantityError(
                'Cannot initialize a stable IsotopeQuantity from activity')
        elif '_init_empty' in kwargs:
            pass
        else:
            raise IsotopeQuantityError('Missing arg for isotope activity')

    def _check_positive_qty(self, val):
        """Check that the quantity value is a nonnegative float or ufloat.

        Raises:
          ValueError: if val is negative
        """

        val *= 1.   # convert to float, or preserve ufloat, as appropriate
        if val < 0:
            raise ValueError(
                'Mass or activity must be a positive quantity: {}'.format(val))
        return val

    @classmethod
    def from_decays(cls, isotope, n_decays, start_time, stop_time):
        """
        Create an IsotopeQuantity from a known number of decays in an interval.

        Args:
          isotope: string or Isotope instance
          n_decays: int or float of the number of decays in the time interval
          start_time: string or datetime of the beginning of the interval
          stop_time: string or datetime of the end of the interval

        Returns:
          an IsotopeQuantity, referenced to start_time

        Raises:
          TypeError: if start_time or stop_time is not a datetime or string
          ValueError: if timestamps are out of order
        """

        obj = cls(isotope, date=start_time, _init_empty=True)

        stop_time = utils.handle_datetime(stop_time)
        duration = (stop_time - obj.ref_date).total_seconds()
        if duration < 0:
            raise ValueError(
                'Start time must precede stop time: {}, {}'.format(
                    start_time, stop_time))
        atoms = float(n_decays) / (1 - np.exp(-obj.decay_const * duration))

        obj.ref_atoms = obj._atoms_from_kwargs(atoms=atoms)
        return obj

    @classmethod
    def from_comparison(cls, isotope_qty1, counts1, interval1,
                        counts2, interval2):
        """Calculate an IsotopeQuantity by comparison with a known sample.

        Assumes the samples are in identical geometry with the detector.

        Args:
          isotope_qty1: an IsotopeQuantity of the known sample
          counts1: net counts measured in the known sample
          interval1: (start_time, stop_time) of the known sample measurement
          counts2: net counts measured in the unknown sample
          interval2: (start_time, stop_time) of the unknown sample measurement

        Returns:
          an IsotopeQuantity of the unknown sample

        Raises:
          IsotopeQuantityError: if intervals are not length 2
          TypeError: if interval elements are not datetimes or date strings
          ValueError: if timestamps are out of order
        """

        norm = decay_normalize(isotope_qty1.isotope, interval1, interval2)
        ratio = (counts2 * norm) / counts1

        return isotope_qty1 * ratio

    # ----------------------------
    #   *_at()
    # ----------------------------

    def atoms_at(self, date):
        """Calculate the number of atoms at a given time.

        Args:
          date: the date to calculate for

        Returns:
          a float of the number of atoms at date

        Raises:
          TypeError: if date is not recognized
        """

        t1 = utils.handle_datetime(date)
        dt = (t1 - self.ref_date).total_seconds()
        return self.ref_atoms * 2**(-dt / self.half_life)

    def bq_at(self, date):
        """Calculate the activity [Bq] at a given time.

        As atoms_at() except for return value.
        """

        return self.atoms_at(date) * self.decay_const

    def uci_at(self, date):
        """Calculate the activity [uCi] at a given time.

        As atoms_at() except for return value.
        """

        return self.bq_at(date) / UCI_TO_BQ

    def g_at(self, date):
        """Calculate the mass [g] at a given time.

        As atoms_at() except for return value.
        """

        return self.atoms_at(date) / N_AV * self.isotope.A

    # ----------------------------
    #   *_now()
    # ----------------------------

    def atoms_now(self):
        """Calculate the number of atoms now.

        Returns:
          a float of the number of atoms at datetime.datetime.now()
        """

        return self.atoms_at(datetime.datetime.now())

    def bq_now(self):
        """Calculate the activity [Bq] now.

        As atoms_now() except for return value.
        """

        return self.bq_at(datetime.datetime.now())

    def uci_now(self):
        """Calculate the activity [uCi] now.

        As atoms_now() except for return value.
        """

        return self.uci_at(datetime.datetime.now())

    def g_now(self):
        """Calculate the mass [g] now.

        As atoms_now() except for return value.
        """

        return self.g_at(datetime.datetime.now())

    # ----------------------------
    #   *_from()
    # ----------------------------

    def decays_from(self, start_time, stop_time):
        """The expected number of decays from start_time to stop_time.

        Args:
          start_time: a string or datetime.datetime object
          stop_time: a string or datetime.datetime object

        Returns:
          a float of the number of decays in the time interval

        Raises:
          TypeError: if start_time or stop_time is not recognized
        """

        return self.atoms_at(start_time) - self.atoms_at(stop_time)

    def bq_from(self, start_time, stop_time):
        """Average activity [Bq] from start_time to stop_time.

        As decays_from() except for return value.
        """

        t0 = utils.handle_datetime(start_time, error_name='start_time')
        t1 = utils.handle_datetime(stop_time, error_name='stop_time')

        return self.decays_from(t0, t1) / (t1 - t0).total_seconds()

    def uci_from(self, start_time, stop_time):
        """Average activity [uCi] from start_time to stop_time.

        As decays_from() except for return value.
        """

        return self.bq_from(start_time, stop_time) / UCI_TO_BQ

    # ----------------------------
    #   *_during()
    # ----------------------------

    def decays_during(self, spec):
        """Calculate the expected number of decays during a measured spectrum.

        Args:
          spec: a Spectrum object containing start_time and stop_time

        Returns:
          a float of the number of decays during the acquisition of spec

        Raises:
          TypeError: if spec does not have start_time or stop_time defined
        """

        return self.decays_from(spec.start_time, spec.stop_time)

    def bq_during(self, spec):
        """Average activity [Bq] during the spectrum.

        As decays_during(), except for return value.
        """

        return self.bq_from(spec.start_time, spec.stop_time)

    def uci_during(self, spec):
        """Average activity [uCi] during the spectrum.

        As decays_during(), except for return value.
        """

        return self.uci_from(spec.start_time, spec.stop_time)

    # ----------------------------
    #   (other)
    # ----------------------------

    def time_when(self, **kwargs):
        """Calculate the date/time when the mass/activity is a given value.

        Args (specify one):
          atoms: number of atoms
          bq: activity [Bq]
          uci: activity [uCi]
          g: mass [g]

        Returns:
          a datetime.datetime of the moment when the mass/activity equals the
            specified input

        Raises:
          IsotopeQuantityError: if isotope is stable
        """

        if self.is_stable:
            raise IsotopeQuantityError(
                'Cannot calculate time_when for stable isotope')

        target = self._atoms_from_kwargs(**kwargs)
        dt = -self.half_life * np.log2(target / self.ref_atoms)
        return self.ref_date + datetime.timedelta(seconds=dt)

    def __str__(self):
        """Return a string representation.

        Shows grams if isotope is stable, otherwise Bq.
        """

        if self.isotope.is_stable:
            s = '{} g of {}'.format(self.g_at(self.ref_date), self.isotope)
        else:
            s = '{} Bq of {} (at {})'.format(
                self.bq_at(self.ref_date), self.isotope, self.ref_date)
        return s

    def __mul__(self, other):
        """Multiply the quantity"""

        return self._mul_div(other, div=False)

    def __div__(self, other):
        """Divide the quantity"""

        return self._mul_div(other, div=True)

    def __truediv__(self, other):
        """Divide the quantity (python 3)"""

        return self._mul_div(other, div=True)

    def _mul_div(self, other, div=False):
        """Multiply or divide the quantity.

        Args:
          other: a scalar to multiply/divide by
          div: a bool, True if dividing, False if multiplying

        Returns:
          a new IsotopeQuantity, same reference date, scaled quantity
        """

        if div:
            factor = 1 / float(other)
        else:
            factor = float(other)
        return IsotopeQuantity(
            copy.deepcopy(self.isotope), date=self.ref_date,
            atoms=self.ref_atoms * factor)

    def __eq__(self, other):
        """Equality operation"""

        if not isinstance(other, IsotopeQuantity):
            return False
        else:
            return (self.isotope == other.isotope and
                    np.isclose(self.ref_atoms, other.atoms_at(self.ref_date)))


class NeutronIrradiationError(Exception):
    """Exception from NeutronIrradiation class."""

    pass


class NeutronIrradiation(object):
    """Represents an irradiation period with thermal neutrons.

    Data attributes:
      start_time: beginning of irradiation
      stop_time: end of irradiation
      duration: number of seconds of irradiation
      n_cm2_s: neutron flux (if duration is nonzero)
      n_cm2: neutron fluence

    Methods:
      activate: Calculate an IsotopeQuantity from before or after irradiation
    """

    def __init__(self, start_time, stop_time, n_cm2=None, n_cm2_s=None):
        """Initialize.

        Either n_cm2 or n_cm2_s is a required input.

        Args:
          start_time: datetime or date string representing start of irradiation
          stop_time: datetime or date string representing end of irradiation
          n_cm2: the total fluence of neutrons over the irradiation.
          n_cm2_s: the flux of neutrons during the irradiation.

        Raises:
          TypeError: if timestamps are not parseable
          ValueError: if timestamps are out of order,
            or if flux/fluence not specified,
            or if flux and fluence both specified
        """

        self.start_time = utils.handle_datetime(
            start_time, error_name='NeutronIrradiation start_time')
        self.stop_time = utils.handle_datetime(
            stop_time, error_name='NeutronIrradiation stop_time')
        if self.stop_time < self.start_time:
            raise ValueError('Timestamps out of order: {}, {}'.format(
                self.start_time, self.stop_time))
        self.duration = (self.stop_time - self.start_time).total_seconds()

        if not ((n_cm2 is None) ^ (n_cm2_s is None)):
            raise ValueError('Must specify either n_cm2 or n_cm2_s, not both')
        elif n_cm2 is None:
            self.n_cm2_s = n_cm2_s
            self.n_cm2 = n_cm2_s * self.duration
        elif n_cm2_s is None and self.duration > 0:
            self.n_cm2_s = n_cm2 / self.duration
            self.n_cm2 = n_cm2
        else:
            self.n_cm2_s = None
            self.n_cm2 = n_cm2

    def __str__(self):
        """Return a string representation.

        Shows flux if duration is nonzero, otherwise shows fluence.
        """

        if self.duration == 0:
            return '{} neutrons/cm2 at {}'.format(self.n_cm2, self.start_time)
        else:
            return '{} n/cm2/s from {} to {}'.format(
                self.n_cm2_s, self.start_time, self.stop_time)

    def activate(self, barns, initial, activated):
        """
        Calculate an IsotopeQuantity from before or after a neutron activation.

        For a forward calculation (known initial quantity, to calculate the
        activated result), specify an initial IsotopeQuantity and an
        activated Isotope.

        For a backward calculation (known activation product, to calculate the
        initial quantity), specify an initial Isotope and an activated
        IsotopeQuantity.

        Forward equations:
          A1 = phi * sigma * N0 * (1 - exp(-lambda * t_irr))
          A1 = n * sigma * N0 * lambda
        Backward equations:
          N0 = A1 / (phi * sigma * (1 - exp(-lambda * t_irr)))
          N0 = A1 / (n * sigma * lambda)

        in all equations:
          A1 = activated activity [Bq] at end of irradiation,
          phi = flux [neutrons/cm2/s],
          sigma = activation cross-section [cm2],
          N0 = number of atoms of initial isotope,
          lambda = activity coefficient of activated isotope [1/s],
          t_irr = duration of irradiation [s]
          n = fluence of zero-duration irradiation [neutrons/cm2],

        Args:
          barns: cross section for activation [barns = 1e-24 cm^2]
          initial: the isotope being activated, an IsotopeQuantity or Isotope.
            Specify an IsotopeQuantity if the initial quantity is known.
            Specify an Isotope if the initial quantity is unknown
          activated: the activated isotope, an IsotopeQuantity or Isotope.
            Specify an IsotopeQuantity if the activated quantity is known.
            Specify an Isotope if the activated quantity is unknown

        Returns:
          an IsotopeQuantity, corresponding to either the initial isotope or
            the activated isotope, depending on which quantity was input

        Raises:
          NeutronIrradiationError: if initial and activated are overspecified
            or underspecified
          TypeError: if initial and activated are not Isotope or
            IsotopeQuantity objects
        """

        if (isinstance(initial, IsotopeQuantity) and
                isinstance(activated, IsotopeQuantity)):
            raise NeutronIrradiationError(
                "Two IsotopeQuantity's in args, nothing left to calculate!" +
                'Args: {}, {}'.format(initial, activated))
        elif (isinstance(initial, IsotopeQuantity) and
              isinstance(activated, Isotope)):
            forward = True
        elif (isinstance(initial, Isotope) and
              isinstance(activated, IsotopeQuantity)):
            forward = False
        elif isinstance(initial, Isotope) and isinstance(activated, Isotope):
            raise NeutronIrradiationError(
                'No IsotopeQuantity specified, not enough data. ' +
                'Args: {}, {}'.format(initial, activated))
        else:
            raise TypeError(
                'Input args should be Isotope or IsotopeQuantity objects: ' +
                '{}, {}'.format(initial, activated))

        if not initial.is_stable:
            raise NotImplementedError(
                'Activation not implemented for a radioactive initial isotope')

        cross_section = barns * 1.0e-24

        if forward:
            if self.duration == 0:
                activated_bq = (
                    self.n_cm2 * cross_section *
                    initial.atoms_at(self.stop_time) *
                    activated.decay_const)
            else:
                activated_bq = (
                    self.n_cm2_s * cross_section *
                    initial.atoms_at(self.stop_time) *
                    (1 - np.exp(-activated.decay_const * self.duration))
                )
            return IsotopeQuantity(activated,
                                   date=self.stop_time, bq=activated_bq)
        else:
            if self.duration == 0:
                initial_atoms = (
                    activated.bq_at(self.stop_time) /
                    (self.n_cm2 * cross_section * activated.decay_const))
            else:
                initial_atoms = (
                    activated.bq_at(self.stop_time) /
                    (self.n_cm2_s * cross_section * (1 - np.exp(
                        -activated.decay_const * self.duration))))
            return IsotopeQuantity(initial,
                                   date=self.start_time, atoms=initial_atoms)


def decay_normalize(isotope, interval1, interval2):
    """Calculate the ratio to normalize decays between time intervals.

    If interval2 averages 1 Bq, what is interval1's average?

    Args:
      isotope: Isotope object or string of the isotope that is decaying
      interval1: (start_time, stop_time) in datetimes or strings
      interval2: (start_time, stop_time) in datetimes or strings

    Returns:
      ratio of (expected decays in interval1) / (expected decays in interval2).
        In other words, multiply measured counts in interval2 by this ratio
        to get the expected counts in interval1.

    Raises:
      IsotopeQuantityError: if intervals are not of length 2
      ValueError: if timestamps are out of order
      TypeError: if timestamps are not parseable, or isotope is not an Isotope
    """

    isotope = handle_isotope(isotope, error_name='decay_normalize')
    if len(interval1) != 2:
        raise IsotopeQuantityError(
            'interval1 should be length 2: {}'.format(interval1))
    elif len(interval2) != 2:
        raise IsotopeQuantityError(
            'interval2 should be length 2: {}'.format(interval2))
    start1 = utils.handle_datetime(interval1[0], error_name='decay_normalize')
    stop1 = utils.handle_datetime(interval1[1], error_name='decay_normalize')
    start2 = utils.handle_datetime(interval2[0], error_name='decay_normalize')
    stop2 = utils.handle_datetime(interval2[1], error_name='decay_normalize')
    if stop1 < start1:
        raise ValueError('Timestamps in interval1 out of order: {}, {}'.format(
            start1, stop1))
    elif stop2 < start2:
        raise ValueError('Timestamps in interval2 out of order: {}, {}'.format(
            start2, stop2))

    # TODO base this on countrate, not counts

    iq = IsotopeQuantity.from_decays(isotope, 1.0, start2, stop2)
    return iq.decays_from(start1, stop1)


def decay_normalize_spectra(isotope, spec1, spec2):
    """Calculate the ratio to normalize decays between measurements.

    If spec2 averages 1 Bq, what is spec1's average?

    Args:
      isotope: Isotope object or string of the isotope that is decaying
      spec1: Spectrum object with start_time, stop_time
      spec2: Spectrum object with start_time, stop_time

    Returns:
      ratio of (expected decays during spec1) / (expected decays during spec2).
        In other words, multiply measured counts in spec2 by this ratio
        to get the expected counts in spec1.

    Raises:
      TypeError: if isotope is not an Isotope instance
      AttributeError: if spec1 or spec2 do not have start_time and stop_time
    """

    return decay_normalize(isotope,
                           (spec1.start_time, spec1.stop_time),
                           (spec2.start_time, spec2.stop_time))



"""Base class for spectrum file parsers."""

from __future__ import print_function
import os
from copy import deepcopy
import datetime
import numpy as np
from uncertainties import UFloat, unumpy
from .. import parsers
from .utils import handle_uncs, handle_datetime, bin_centers_from_edges, EPS
from .rebin import rebin
from . import plotting
import warnings


class SpectrumError(Exception):
    """Exception raised by Spectrum."""

    pass


class SpectrumWarning(UserWarning):
    """Warnings displayed by Spectrum."""

    pass


class UncalibratedError(SpectrumError):
    """Raised when an uncalibrated spectrum is treated as calibrated."""

    pass


class Spectrum(object):
    """
    Represents an energy spectrum.

    Initialize a Spectrum directly, or with Spectrum.from_file(filename), or
    with Spectrum.from_listmode(listmode_data).

    Note on livetime:
      A livetime of None is the default for a spectrum, and indicates a
        missing, unknown livetime or not meaningful quantity.
      A spectrum may be initialized with a livetime value. However, any
        operation that produces a CPS-based spectrum (such as a spectrum
        subtraction) will discard the livetime and set it to None.
      Operations that produce a counts-based spectrum may or may not preserve a
        livetime value (for example, the sum of two spectra has a livetime
        equal to the sum of the two livetimes; but a scalar multiplication or
        division results in a livetime of None).

    Data Attributes:
      bin_edges_raw: np.array of raw/uncalibrated bin edges
      bin_edges_kev: np.array of energy bin edges, if calibrated
      livetime: int or float of livetime, in seconds. See note above
      realtime: int or float of realtime, in seconds. May be None
      infilename: the filename the spectrum was loaded from, if applicable
      start_time: a datetime.datetime object representing the acquisition start
      stop_time: a datetime.datetime object representing the acquisition end

    Properties (read-only):
      counts: counts in each bin, with uncertainty
      counts_vals: counts in each bin, no uncertainty
      counts_uncs: uncertainties on counts in each bin
      cps: counts per second in each bin, with uncertainty
      cps_vals: counts per second in each bin, no uncertainty
      cps_uncs: uncertainties on counts per second in each bin
      cpskev: CPS/keV in each bin, with uncertainty
      cpskev_vals: CPS/keV in each bin, no uncertainty
      cpskev_uncs: uncertainties on CPS/keV in each bin
      bins: np.array of bin index as integers
      channels: np.array of bin index as integers (deprecated, raises a
        DeprecationWarning)
      is_calibrated: bool indicating calibration status
      energies_kev: np.array of energy bin centers, if calibrated (deprecated,
        raises a DeprecationWarning)
      bin_centers_raw: np.array of raw bin centers
      bin_centers_kev: np.array of energy bin centers, if calibrated
      bin_widths: np.array of energy bin widths, if calibrated (deprecated,
        raises a DeprecationWarning)
      bin_widths_raw: np.array of raw bin widths
      bin_widths_kev: np.array of energy bin widths, if calibrated

    Methods:
      apply_calibration: use an EnergyCal object to calibrate this spectrum
      calibrate_like: copy the calibrated bin edges from another spectrum
      rm_calibration: remove the calibrated bin edges
      combine_bins: make a new Spectrum with counts combined into bigger bins
      downsample: make a new Spectrum, downsampled from this one by a factor
      copy: return a deep copy of this Spectrum object
    """

    def __init__(self, counts=None, cps=None, uncs=None, bin_edges_kev=None,
                 bin_edges_raw=None, input_file_object=None, livetime=None,
                 realtime=None, start_time=None, stop_time=None):
        """Initialize the spectrum.

        Either counts or cps must be specified. Other args are optional.
        If start_time, stop_time, and realtime are being provided, only two of
        these should be specified as arguments, and the third will be
        calculated from the other two.

        See note on livetime in class docstring.

        Args:
          counts: counts per bin. array-like of ints, floats or UFloats, if
            uncs is not provided all values must be positive
          cps: counts per second per bin. array-like of floats or UFloats
          uncs (optional): iterable of uncertainty on the counts for each bin.
            If counts is given, is NOT a UFloat type, and uncs is not given,
            the uncertainties are assumed to be sqrt(N), with a minimum
            uncertainty of 1 (e.g. for 0 counts).
            If cps is given, uncs defaults to an array of np.nan
          bin_edges_kev (optional): an iterable of bin edge energies
            If not none, should have length of (len(counts) + 1)
          input_file_object (optional): a parser file object
          livetime (optional): the livetime of the spectrum [s]
            Note that livetime is not preserved through CPS-based operations,
            such as any subtraction, or addition with a CPS-based spectrum.
          realtime (optional): the duration of the acquisition [s]
          start_time (optional): datetime or string representing the
            acquisition start
          stop_time (optional): datetime or string representing the acquisition
            end.

        Raises:
          ValueError: for bin edges not monotonically increasing;
            livetime > realtime; or start_time > stop_time
          SpectrumError: for bad input arguments
          UncertaintiesError: if uncertainties are overspecified or of mixed
            types
          TypeError: for bad type on input args (start_time, stop_time)
        """

        if not (counts is None) ^ (cps is None):
            raise SpectrumError('Must specify one of counts or CPS')

        self._counts = None
        self._cps = None
        self._bin_edges_kev = None
        self._bin_edges_raw = None
        self.livetime = None
        self.realtime = None

        if counts is not None:
            if len(counts) == 0:
                raise SpectrumError('Empty spectrum counts')
            if uncs is None and np.any(np.asarray(counts) < 0):
                raise SpectrumError(
                    'Negative values encountered in counts. Uncertainties ' +
                    'are most likely not Poisson-distributed. Provide uncs ' +
                    'to force initialization.')
            self._counts = handle_uncs(
                counts, uncs, lambda x: np.maximum(np.sqrt(x), 1))
        else:
            if len(cps) == 0:
                raise SpectrumError('Empty spectrum counts')
            self._cps = handle_uncs(cps, uncs, lambda x: np.nan)

        if bin_edges_raw is None and not (counts is None and cps is None):
            bin_edges_raw = np.arange(len(self)+1)
        self.bin_edges_raw = bin_edges_raw
        self.bin_edges_kev = bin_edges_kev

        if livetime is not None:
            self.livetime = float(livetime)

        if realtime is not None:
            self.realtime = float(realtime)
            if self.livetime is not None:
                if self.livetime > self.realtime:
                    raise ValueError(
                        'Livetime ({}) cannot exceed realtime ({})'.format(
                            self.livetime, self.realtime))

        self.start_time = handle_datetime(
            start_time, 'start_time', allow_none=True)
        self.stop_time = handle_datetime(
            stop_time, 'stop_time', allow_none=True)

        if (self.realtime is not None and
                self.stop_time is not None and
                self.start_time is not None):
            raise SpectrumError(
                'Specify no more than 2 out of 3 args: ' +
                'realtime, stop_time, start_time')
        elif self.start_time is not None and self.stop_time is not None:
            if self.start_time > self.stop_time:
                raise ValueError(
                    'Stop time ({}) must be after start time ({})'.format(
                        self.start_time, self.stop_time))
            self.realtime = (self.stop_time - self.start_time).total_seconds()
        elif self.start_time is not None and self.realtime is not None:
            self.stop_time = self.start_time + datetime.timedelta(
                seconds=self.realtime)
        elif self.realtime is not None and self.stop_time is not None:
            self.start_time = self.stop_time - datetime.timedelta(
                seconds=self.realtime)

        self._infileobject = input_file_object
        if input_file_object is not None:
            self.infilename = input_file_object.filename
            if self.livetime is None:
                self.livetime = input_file_object.livetime
            if self.realtime is None:
                self.realtime = input_file_object.realtime
            if self.start_time is None:
                self.start_time = input_file_object.collection_start
            if self.stop_time is None:
                self.stop_time = input_file_object.collection_stop
        else:
            self.infilename = None
        # These two lines make sure operators between a Spectrum
        # and a numpy arrays are forbidden and cause a TypeError
        self.__array_ufunc__ = None
        self.__array_priority__ = 1

    def __str__(self):
        lines = ['becquerel.Spectrum']
        ltups = []
        for k in ['start_time', 'stop_time', 'realtime', 'livetime',
                  'is_calibrated']:
            ltups.append((k, getattr(self, k)))
        ltups.append(('num_bins', len(self.bin_indices)))
        if self._counts is None:
            ltups.append(('gross_counts', None))
        else:
            ltups.append(('gross_counts', self.counts.sum()))
        try:
            ltups.append(('gross_cps', self.cps.sum()))
        except SpectrumError:
            ltups.append(('gross_cps', None))
        if hasattr(self, 'infilename'):
            ltups.append(('filename', self.infilename))
        else:
            ltups.append(('filename', None))
        for lt in ltups:
            lines.append('    {:15} {}'.format(
                '{}:'.format(lt[0]),
                lt[1]))
        return '\n'.join(lines)

    __repr__ = __str__

    @property
    def counts(self):
        """Counts in each bin, with uncertainty.

        If cps is defined, counts is calculated from cps and livetime.
        Otherwise, it is an independent data property.

        Raises:
          SpectrumError: if cps is defined, but not livetime

        Returns:
          an np.array of uncertainties.ufloats
        """

        if self._counts is not None:
            return self._counts
        else:
            try:
                return self.cps * self.livetime
            except TypeError:
                raise SpectrumError(
                    'Unknown livetime; cannot calculate counts from CPS')

    @property
    def counts_vals(self):
        """Counts in each bin, no uncertainties.

        Returns:
          an np.array of floats
        """

        return unumpy.nominal_values(self.counts)

    @property
    def counts_uncs(self):
        """Uncertainties on the counts in each bin.

        Returns:
          an np.array of floats
        """

        return unumpy.std_devs(self.counts)

    @property
    def cps(self):
        """Counts per second in each bin, with uncertainty.

        If counts is defined, cps is calculated from counts and livetime.
        Otherwise, it is an independent data property.

        Raises:
          SpectrumError: if counts is defined, but not livetime

        Returns:
          an np.array of uncertainties.ufloats
        """

        if self._cps is not None:
            return self._cps
        else:
            try:
                return self.counts / self.livetime
            except TypeError:
                raise SpectrumError(
                    'Unknown livetime; cannot calculate CPS from counts')

    @property
    def cps_vals(self):
        """Counts per second in each bin, no uncertainties.

        Returns:
          an np.array of floats
        """

        return unumpy.nominal_values(self.cps)

    @property
    def cps_uncs(self):
        """Uncertainties on the counts per second in each bin.

        Returns:
          an np.array of floats
        """

        return unumpy.std_devs(self.cps)

    @property
    def cpskev(self):
        """Counts per second per keV in each bin, with uncertainty.

        Raises:
          SpectrumError: if cps is not defined due to missing livetime
          UncalibratedError: if bin edges (and thus bin widths) are not defined

        Returns:
          an np.array of uncertainties.ufloats
        """

        return self.cps / self.bin_widths_kev

    @property
    def cpskev_vals(self):
        """Counts per second per keV in each bin, no uncertainties.

        Returns:
          an np.array of floats
        """

        return unumpy.nominal_values(self.cpskev)

    @property
    def cpskev_uncs(self):
        """Uncertainties on the counts per second per keV in each bin.

        Returns:
          an np.array of floats
        """

        return unumpy.std_devs(self.cpskev)

    @property
    def bin_indices(self):
        """Bin indices.

        Returns:
          np.array of int's from 0 to (len(self.counts) - 1)
        """

        return np.arange(len(self), dtype=int)

    @property
    def channels(self):
        """Alias for bin_indices. Raises a DeprecationWarning, as it will be
        removed in a future release.

        Returns:
          np.array of int's from 0 to (len(self.counts) - 1)
        """

        warnings.warn('channels is deprecated terminology and will be removed '
                      'in a future release. Use bin_indices instead.',
                      DeprecationWarning)
        return np.arange(len(self), dtype=int)

    @property
    def bin_centers_raw(self):
        """Convenience function for accessing the raw values of bin centers.

        Returns:
          np.array of floats, same length as self.counts
        """

        return bin_centers_from_edges(self.bin_edges_raw)

    @property
    def bin_widths_raw(self):
        """The width of each bin, in raw values.

        Returns:
          np.array of floats, same length as self.counts
        """

        return np.diff(self.bin_edges_raw)

    @property
    def bin_centers_kev(self):
        """Convenience function for accessing the energies of bin centers.

        Returns:
          np.array of floats, same length as self.counts

        Raises:
          UncalibratedError: if spectrum is not calibrated
        """

        if not self.is_calibrated:
            raise UncalibratedError('Spectrum is not calibrated')
        else:
            return bin_centers_from_edges(self.bin_edges_kev)

    @property
    def energies_kev(self):
        """Convenience function for accessing the energies of bin centers.

        Returns:
          np.array of floats, same length as self.counts

        Raises:
          UncalibratedError: if spectrum is not calibrated
          DeprecationWarning: since it will be removed in favor of
            bin_centers_kev in a future release
        """

        warnings.warn('energies_kev is deprecated and will be removed in a '
                      'future release. Use bin_centers_kev instead.',
                      DeprecationWarning)

        if not self.is_calibrated:
            raise UncalibratedError('Spectrum is not calibrated')
        else:
            return bin_centers_from_edges(self.bin_edges_kev)

    @property
    def bin_widths_kev(self):
        """The width of each bin, in keV.

        Returns:
          np.array of floats, same length as self.counts

        Raises:
          UncalibratedError: if spectrum is not calibrated
        """

        if not self.is_calibrated:
            raise UncalibratedError('Spectrum is not calibrated')
        else:
            return np.diff(self.bin_edges_kev)

    @property
    def bin_widths(self):
        """The width of each bin, in keV.

        Returns:
          np.array of floats, same length as self.counts

        Raises:
          UncalibratedError: if spectrum is not calibrated
          DeprecationWarning: since it will be removed in favor of
            bin_widths_kev in a future release
        """

        warnings.warn('bin_widths is deprecated and will be removed in a '
                      'future release. Use bin_widths_kev (or bin_widths_raw) '
                      'instead.', DeprecationWarning)

        if not self.is_calibrated:
            raise UncalibratedError('Spectrum is not calibrated')
        else:
            return np.diff(self.bin_edges_kev)

    @property
    def is_calibrated(self):
        """Is the spectrum calibrated?

        Returns:
          bool, True if spectrum has defined energy bin edges. False otherwise
        """

        return self.bin_edges_kev is not None

    @property
    def bin_edges_kev(self):
        """Get the bin edge energies of a spectrum

        Returns:
          np.array of floats or None
        """

        return self._bin_edges_kev

    @bin_edges_kev.setter
    def bin_edges_kev(self, bin_edges_kev):
        """Set the bin edge energies of a spectrum

        Args:
          bin_edges_kev: an iterable of bin edge energies
            If not None, should have length of (len(counts) + 1)

        Raises:
          SpectrumError: for bad input length
        """
        if bin_edges_kev is None:
            self._bin_edges_kev = None
        elif len(bin_edges_kev) != len(self) + 1:
            raise SpectrumError('Bad length of bin edges vector')
        elif np.any(np.diff(bin_edges_kev) <= 0):
            raise ValueError(
                'Bin edge energies must be strictly increasing')
        else:
            self._bin_edges_kev = np.array(bin_edges_kev, dtype=float)

    @property
    def bin_edges_raw(self):
        """Get the raw bin edges of a spectrum

        Returns:
          np.array of floats or None
        """

        return self._bin_edges_raw

    @bin_edges_raw.setter
    def bin_edges_raw(self, bin_edges_raw):
        """Set the raw bin edges of a spectrum

        Args:
          bin_edges_raw: an iterable of raw bin edges
            If not None, should have length of (len(counts) + 1)

        Raises:
          SpectrumError: for bad input length
        """
        if bin_edges_raw is None:
            self._bin_edges_raw = None
        elif len(bin_edges_raw) != len(self) + 1:
            raise SpectrumError('Bad length of bin edges vector')
        elif np.any(np.diff(bin_edges_raw) <= 0):
            raise ValueError(
                'Raw bin edges must be strictly increasing')
        else:
            self._bin_edges_raw = np.array(bin_edges_raw, dtype=float)

    @classmethod
    def from_file(cls, infilename):
        """Construct a Spectrum object from a filename.

        Args:
          infilename: a string representing the path to a parsable file

        Returns:
          A Spectrum object

        Raises:
          AssertionError: for a bad filename  # TODO make this an IOError
        """

        spect_file_obj = _get_file_object(infilename)

        kwargs = {'counts': spect_file_obj.data,
                  'input_file_object': spect_file_obj,
                  'bin_edges_kev': spect_file_obj.bin_edges_kev}

        # TODO Get more attributes from self.infileobj

        return cls(**kwargs)

    @classmethod
    def from_listmode(cls, listmode_data, bins=None, xmin=None, xmax=None,
                      **kwargs):
        """Construct a Spectrum object (specifically the `bin_edges_raw` and
        `counts` of a histogram) from an array of listmode data. It is left to
        the user to set kwargs realtime, livetime, etc, rather than trying to
        cover all use cases here.

        Args:
          listmode_data: the array containing the listmode data, e.g., the ADC
            values triggered or pulse integral values recorded
          bins: integer number of bins OR array of bin_edges, following the
            numpy.histogram style (array of all low edges and last up edge)
          xmin: minimum x of histogram; equals bin_edges[0] if int # of bins
          xmax: maximum x of histogram; equals bin_edges[-1] if int # of bins

        Returns:
          A Spectrum object

        Raises:
          AssertionError: no listmode_data, or xmin >= xmax, or nbins < 1
        """

        assert len(listmode_data) > 0

        if xmin is None:
            xmin = 0
        if xmax is None:
            xmax = np.ceil(max(listmode_data))
        if bins is None:
            bins = np.arange(xmin, xmax + 1, dtype=np.int)

        assert xmin < xmax
        if isinstance(bins, int):
            assert bins > 0
        else:
            assert len(bins) > 1

        bin_counts, bin_edges = np.histogram(listmode_data,
                                             bins=bins,
                                             range=(xmin, xmax))

        kwargs['counts'] = bin_counts
        kwargs['bin_edges_raw'] = bin_edges

        return cls(**kwargs)

    def copy(self):
        """Make a deep copy of this Spectrum object.

        Returns:
          a Spectrum object identical to this one
        """

        return deepcopy(self)

    def __len__(self):
        """The number of bins in the spectrum.

        Returns:
          an int
        """

        try:
            return len(self.counts)
        except SpectrumError:
            return len(self.cps)

    def __add__(self, other):
        """Add spectra together.

        The livetimes sum (if they exist) and the resulting spectrum
        is still Poisson-distributed.

        The two spectra may both be uncalibrated, or both be calibrated
        with the same energy calibration.

        Args:
          other: another Spectrum object to add counts from

        Raises:
          TypeError: if other is not a Spectrum
          SpectrumError: if spectra are different lengths,
            if only one is calibrated or if spectra are not both
            counts/CPS-based, respectively.
          NotImplementedError: if spectra are calibrated differently

        Returns:
          a summed Spectrum object
        """

        self._add_sub_error_checking(other)
        if (self._counts is None) ^ (other._counts is None):
            raise SpectrumError(
                'Addition of counts-based and CPS-based spectra is ' +
                'ambiguous, use Spectrum(counts=specA.counts+specB.counts) ' +
                'or Spectrum(cps=specA.cps+specB.cps) instead.')

        if self._counts is not None and other._counts is not None:
            kwargs = {'counts': self.counts + other.counts}
            if self.livetime and other.livetime:
                kwargs['livetime'] = self.livetime + other.livetime
            else:
                warnings.warn('Addition of counts with missing livetimes, ' +
                              'livetime was set to None.', SpectrumWarning)
        else:
            kwargs = {'cps': self.cps + other.cps}

        if self.is_calibrated and other.is_calibrated:
            spect_obj = Spectrum(bin_edges_kev=self.bin_edges_kev, **kwargs)
        else:
            spect_obj = Spectrum(bin_edges_raw=self.bin_edges_raw, **kwargs)
        return spect_obj

    def __sub__(self, other):
        """Normalize spectra (if possible) and subtract.

        The resulting spectrum does not have a meaningful livetime or
        counts vector, and is NOT Poisson-distributed.

        The two spectra may both be uncalibrated, or both be calibrated
        with the same energy calibration.

        Args:
          other: another Spectrum object, to normalize and subtract

        Raises:
          TypeError: if other is not a Spectrum
          SpectrumError: if spectra are different lengths or
            if only one is calibrated.
          NotImplementedError: if spectra are calibrated differently

        Warns:
          SpectrumWarning: If both spectrum are counts-based, or if one
            of them has been converted to CPS during the operation.

        Returns:
          a subtracted Spectrum object
        """

        self._add_sub_error_checking(other)
        try:
            kwargs = {'cps': self.cps - other.cps}
            if (self._cps is None) or (other._cps is None):
                warnings.warn('Subtraction of counts-based specta, spectra ' +
                              'have been converted to CPS', SpectrumWarning)
        except SpectrumError:
            try:
                kwargs = {'counts': self.counts_vals - other.counts_vals}
                kwargs['uncs'] = [np.nan]*len(self)
                warnings.warn('Subtraction of counts-based spectra, ' +
                              'livetimes have been ignored.', SpectrumWarning)
            except SpectrumError:
                raise SpectrumError(
                    'Subtraction of counts and CPS-based spectra without' +
                    'livetimes not possible')

        if self.is_calibrated and other.is_calibrated:
            spect_obj = Spectrum(bin_edges_kev=self.bin_edges_kev, **kwargs)
        else:
            spect_obj = Spectrum(bin_edges_raw=self.bin_edges_raw, **kwargs)
        return spect_obj

    def _add_sub_error_checking(self, other):
        """Handle errors for spectra addition or subtraction.

        Args:
          other: a spectrum

        Raises:
          TypeError: if other is not a Spectrum
          SpectrumError: if spectra are different lengths or
            if only one is calibrated.
          NotImplementedError: if spectra are calibrated differently
        """

        if not isinstance(other, Spectrum):
            raise TypeError(
                'Spectrum addition/subtraction must involve a Spectrum object')
        if len(self) != len(other):
            raise SpectrumError(
                'Cannot add/subtract spectra of different lengths')
        if self.is_calibrated ^ other.is_calibrated:
            raise SpectrumError(
                'Cannot add/subtract uncalibrated spectrum to/from a ' +
                'calibrated spectrum. If both have the same calibration, ' +
                'please use the "calibrate_like" method')
        if self.is_calibrated and other.is_calibrated:
            if not np.all(self.bin_edges_kev == other.bin_edges_kev):
                raise NotImplementedError(
                    'Addition/subtraction for arbitrary calibrated spectra ' +
                    'not implemented')
                # TODO: if both spectra are calibrated but with different
                #   calibrations, should one be rebinned to match?
        if not self.is_calibrated and not other.is_calibrated:
            if not np.all(self.bin_edges_raw == other.bin_edges_raw):
                raise NotImplementedError(
                    'Addition/subtraction for arbitrary uncalibrated ' +
                    'spectra not implemented')

    def __mul__(self, other):
        """Return a new Spectrum object with counts (or CPS) scaled up.

        Args:
          factor: factor to multiply by. May be a ufloat.

        Raises:
          TypeError: if factor is not a scalar value
          SpectrumError: if factor is 0 or infinite

        Returns:
          a new Spectrum object
        """

        return self._mul_div(other, div=False)

    # This line adds the right multiplication
    __rmul__ = __mul__

    def __div__(self, other):
        """Return a new Spectrum object with counts (or CPS) scaled down.

        Args:
          factor: factor to divide by. May be a ufloat.

        Raises:
          TypeError: if factor is not a scalar value
          SpectrumError: if factor is 0 or infinite

        Returns:
          a new Spectrum object
        """

        return self._mul_div(other, div=True)

    # This line adds true division
    __truediv__ = __div__

    def _mul_div(self, scaling_factor, div=False):
        """Multiply or divide a spectrum by a scalar. Handle errors.

        Raises:
          TypeError: if factor is not a scalar value
          ValueError: if factor is 0 or infinite

        Returns:
          a new Spectrum object
        """

        if not isinstance(scaling_factor, UFloat):
            try:
                scaling_factor = float(scaling_factor)
            except (TypeError, ValueError):
                raise TypeError(
                    'Spectrum must be multiplied/divided by a scalar')
            if (scaling_factor == 0 or
                    np.isinf(scaling_factor) or
                    np.isnan(scaling_factor)):
                raise ValueError(
                    'Scaling factor must be nonzero and finite')
        else:
            if (scaling_factor.nominal_value == 0 or
                    np.isinf(scaling_factor.nominal_value) or
                    np.isnan(scaling_factor.nominal_value)):
                raise ValueError(
                    'Scaling factor must be nonzero and finite')
        if div:
            multiplier = 1 / scaling_factor
        else:
            multiplier = scaling_factor

        if self._counts is not None:
            data_arg = {'counts': self.counts * multiplier}
        else:
            data_arg = {'cps': self.cps * multiplier}

        if self.is_calibrated:
            spect_obj = Spectrum(bin_edges_kev=self.bin_edges_kev, **data_arg)
        else:
            spect_obj = Spectrum(bin_edges_raw=self.bin_edges_raw, **data_arg)
        return spect_obj

    def downsample(self, f, handle_livetime=None):
        """Downsample counts and create a new spectrum.

        The spectrum is resampled from a binomial distribution. Each count in
        the spectrum is preserved (with a probability of 1/f) or discarded,
        resulting in a new spectrum with fewer counts, in which Poisson
        statistics are preserved (unlike a scalar division operation).

        Note, it is not possible to downsample a CPS-based spectrum because
        such a spectrum does not have "counts" to downsample.

        Args:
          f: factor by which to downsample. Must be greater than 1.
          handle_livetime (optional): Possible values:
            None (default): the resulting spectrum has livetime = None.
            'preserve': the resulting spectrum has the same livetime.
            'reduce': the resulting spectrum has livetime reduced by the
              downsampling factor.

        Raises:
          SpectrumError: if this spectrum is CPS-based
          ValueError: if f < 1, or if handle_livetime is an illegal value

        Returns:
          a new Spectrum instance, downsampled from this spectrum
        """

        if self._counts is None:
            raise SpectrumError('Cannot downsample from CPS')
        if f < 1:
            raise ValueError('Cannot upsample a spectrum; f must be > 1')

        if handle_livetime is None:
            new_livetime = None
        elif handle_livetime.lower() == 'preserve':
            new_livetime = self.livetime
        elif handle_livetime.lower() == 'reduce':
            new_livetime = self.livetime / f
        else:
            raise ValueError('Illegal value for handle_livetime: {}'.format(
                handle_livetime))

        # TODO handle uncertainty?
        old_counts = self.counts_vals.astype(int)
        new_counts = np.random.binomial(old_counts, 1. / f)

        if self.is_calibrated:
            return Spectrum(counts=new_counts,
                            bin_edges_kev=self.bin_edges_kev,
                            livetime=new_livetime)
        else:
            return Spectrum(counts=new_counts,
                            bin_edges_raw=self.bin_edges_raw,
                            livetime=new_livetime)

    def has_uniform_bins(self, use_kev=True):
        """Test whether the Spectrum has uniform binning.

        This is possibly the best way to test for uniform binning within float
        precision, after https://github.com/0cjs/py-allthesame. It may also be
        useful to make this a member, not a method, that gets set at Spectrum
        initialization.

        Args:
          use_kev: check bin_edges_kev if True, otherwise bin_edges_raw. This
            is a rather unfortunate workaround necessary when keeping both the
            raw and kev versions of the bin_edges

        Raises:
          SpectrumError: if use_kev=True but Spectrum is not calibrated

        Returns:
          True/False if all binwidths are equal to within np.finfo(float).eps
        """

        if use_kev and not self.is_calibrated:
            raise SpectrumError('Cannot access energy bins with an ' +
                                'uncalibrated Spectrum.')

        bin_widths = self.bin_widths_kev if use_kev else self.bin_widths_raw
        iterator = iter(bin_widths)
        x0 = next(iterator, None)
        for x in iterator:
            if abs(x/x0 - 1.0) > EPS:
                return False
        return True

    def find_bin_index(self, x, use_kev=True):
        """Find the Spectrum bin index that contains an x-axis value x.

        If the Spectrum has uniform binning, then we just solve the linear
        equation between bin indices and x-axis values. Otherwise, we use
        searchsorted to bisect for the insertion point where x would fit in a
        list of bin edges, then subtract 1 to get the index of the low edge.

        Args:
          x: value whose bin to find
          use_kev: check bin_edges_kev if True, otherwise bin_edges_raw

        Raises:
          AssertionError: if x is outside the bin edges or equal to up edge
          SpectrumError: if use_kev=True but Spectrum is not calibrated

        Returns:
          The integer bin index that contains x
        """

        if use_kev and not self.is_calibrated:
            raise SpectrumError('Cannot access energy bins with an ' +
                                'uncalibrated Spectrum.')

        bin_edges = self.bin_edges_kev if use_kev else self.bin_edges_raw
        bin_widths = self.bin_widths_kev if use_kev else self.bin_widths_raw

        assert x >= bin_edges[0]
        assert x < bin_edges[-1]

        if self.has_uniform_bins(use_kev=use_kev):
            return int((x - bin_edges[0]) / bin_widths[0])
        else:
            return np.searchsorted(bin_edges, x, 'right') - 1

    def apply_calibration(self, cal):
        """Use an EnergyCal to generate bin edge energies for this spectrum.

        Args:
          cal: an object derived from EnergyCalBase
        """

        self.bin_edges_kev = cal.ch2kev(self.bin_edges_raw)

    def calibrate_like(self, other):
        """Apply another Spectrum object's calibration (bin edges vector).

        Bin edges are copied, so the two spectra do not have the same object
        in memory.

        Args:
          other: spectrum to copy the calibration from

        Raises:
          UncalibratedError: if other Spectrum is not calibrated
        """

        if other.is_calibrated:
            self.bin_edges_kev = other.bin_edges_kev.copy()
        else:
            raise UncalibratedError('Other spectrum is not calibrated')

    def rm_calibration(self):
        """Remove the calibration (if it exists) from this spectrum."""

        self.bin_edges_kev = None

    def combine_bins(self, f):
        """Make a new Spectrum with counts combined into bigger bins.

        If f is not a factor of the number of bins, the counts from the first
        spectrum will be padded with zeros.

        len(new.counts) == np.ceil(float(len(self.counts)) / f)

        Args:
          f: an int representing the number of bins to combine into one

        Returns:
          a Spectrum object with counts from this spectrum, but with
            fewer bins
        """

        f = int(f)
        if self._counts is None:
            key = 'cps'
        else:
            key = 'counts'
        data = getattr(self, key)
        if len(self) % f == 0:
            padded_counts = np.copy(data)
        else:
            pad_len = f - len(self) % f
            pad_counts = unumpy.uarray(np.zeros(pad_len), np.zeros(pad_len))
            padded_counts = np.concatenate((data, pad_counts))
        padded_counts.resize(int(len(padded_counts) / f), f)
        combined_counts = np.sum(padded_counts, axis=1)
        if self.is_calibrated:
            combined_bin_edges = self.bin_edges_kev[::f]
            if combined_bin_edges[-1] != self.bin_edges_kev[-1]:
                combined_bin_edges = np.append(
                    combined_bin_edges, self.bin_edges_kev[-1])
        else:  # TODO: should be able to combine bins
            combined_bin_edges = None

        kwargs = {key: combined_counts,
                  'bin_edges_kev': combined_bin_edges,
                  'input_file_object': self._infileobject,
                  'livetime': self.livetime}
        obj = Spectrum(**kwargs)
        return obj

    def rebin(self, out_edges, method="interpolation", slopes=None,
              zero_pad_warnings=True):
        """
        Spectra rebinning via deterministic or stochastic methods.

        Args:
            out_edges (np.ndarray): an array of the output bin edges
                [num_bins_out]
            method (str): rebinning method
                "interpolation"
                    Deterministic interpolation
                "listmode"
                    Stochastic rebinning via conversion to listmode of energies
            slopes (np.ndarray|None): (optional) an array of input bin slopes
                for quadratic interpolation
                (only applies for "interpolation" method)
                [len(spectrum) + 1]
        zero_pad_warnings (boolean): warn when edge overlap results in
            appending empty bins

        Raises:
            SpectrumError: for bad input arguments

        Returns:
            A new Spectrum object with the rebinned data.
        """
        if self.bin_edges_kev is None:
            raise SpectrumError('Cannot rebin spectrum without energy '
                                'calibration')  # TODO: why not?
        in_spec = self.counts_vals
        if method.lower() == 'listmode':
            if (self._counts is None) and (self.livetime is not None):
                warnings.warn(
                    'Rebinning by listmode method without explicit counts ' +
                    'provided in Spectrum object',
                    SpectrumWarning)
        out_spec = rebin(in_spec, self.bin_edges_kev, out_edges,
                         method=method, slopes=slopes,
                         zero_pad_warnings=zero_pad_warnings)
        return Spectrum(counts=out_spec,
                        uncs=np.sqrt(out_spec),
                        bin_edges_kev=out_edges,
                        input_file_object=self._infileobject,
                        livetime=self.livetime)

    def rebin_like(self, other, zero_pad_warnings=False, **kwargs):
        return self.rebin(other.bin_edges_kev,
                          zero_pad_warnings=zero_pad_warnings, **kwargs)
        # TODO: raw here too?

    def plot(self, *fmt, **kwargs):
        """Plot a spectrum with matplotlib's plot command.

        Args:
          fmt:    matplotlib like plot format string
          xmode:  define what is plotted on x axis ('energy' or 'channel'),
                  defaults to energy if available
          ymode:  define what is plotted on y axis ('counts', 'cps', 'cpskev'),
                  defaults to counts
          xlim:   set x axes limits, if set to 'default' use special scales
          ylim:   set y axes limits, if set to 'default' use special scales
          ax:     matplotlib axes object, if not provided one is created
          yscale: matplotlib scale: 'linear', 'log', 'logit', 'symlog'
          title:  costum plot title
          xlabel: costum xlabel value
          ylabel: costum ylabel value
          emode:  can be 'band' for adding an erroband or 'bars' for adding
                  error bars, default is 'none'. It herits the color from
                  matplotlib plot and can not be configured. For better
                  plotting control use SpectrumPlotter and its errorband and
                  errorbars functions.
          kwargs: arguments that are directly passed to matplotlib's plot
                  command. In addition it is possible to pass linthreshy if
                  ylim='default' and ymode='symlog'

        Returns:
          matplotlib axes object
        """

        emode = 'none'
        alpha = 1
        if 'emode' in kwargs:
            emode = kwargs.pop('emode')
        if 'alpha' in kwargs:
            alpha = kwargs['alpha']

        plotter = plotting.SpectrumPlotter(self, *fmt, **kwargs)
        ax = plotter.plot()
        color = ax.get_lines()[-1].get_color()
        if emode == 'band':
            plotter.errorband(color=color, alpha=alpha*0.5, label='_nolegend_')
        elif emode == 'bars' or emode == 'bar':
            plotter.errorbar(color=color, label='_nolegend_')
        elif emode != 'none':
            raise SpectrumError("Unknown error mode '{}', use 'bars' "
                                "or 'band'".format(emode))
        return ax

    def fill_between(self, **kwargs):
        """Plot a spectrum with matplotlib's fill_between command

        Args:
          xmode:  define what is plotted on x axis ('energy' or 'channel'),
                  defaults to energy if available
          ymode:  define what is plotted on y axis ('counts', 'cps', 'cpskev'),
                  defaults to counts
          xlim:   set x axes limits, if set to 'default' use special scales
          ylim:   set y axes limits, if set to 'default' use special scales
          ax:     matplotlib axes object, if not provided one is created
          yscale: matplotlib scale: 'linear', 'log', 'logit', 'symlog'
          title:  costum plot title
          xlabel: costum xlabel value
          ylabel: costum ylabel value
          kwargs: arguments that are directly passed to matplotlib's
                  fill_between command. In addition it is possible to pass
                  linthreshy if ylim='default' and ymode='symlog'.

        Returns:
          matplotlib axes object
        """

        plotter = plotting.SpectrumPlotter(self, **kwargs)
        return plotter.fill_between()


def _get_file_object(infilename):
    """
    Parse a file and return an object according to its extension.

    Args:
      infilename: a string representing a path to a parsable file

    Raises:
      AssertionError: for a bad filename  # TODO let this be an IOError
      NotImplementedError: for an unparsable file extension
      ...?

    Returns:
      a file object of type SpeFile, SpcFile, or CnfFile
    """

    _, extension = os.path.splitext(infilename)
    if extension.lower() == '.spe':
        return parsers.SpeFile(infilename)
    elif extension.lower() == '.spc':
        return parsers.SpcFile(infilename)
    elif extension.lower() == '.cnf':
        return parsers.CnfFile(infilename)
    else:
        raise NotImplementedError(
            'File type {} can not be read'.format(extension))



"""Base class for spectrum file parsers."""

from __future__ import print_function
import os
import warnings
import numpy as np
from scipy.interpolate import interp1d
warnings.simplefilter('always', DeprecationWarning)


class SpectrumFileParsingWarning(UserWarning):
    """Warnings displayed by SpectrumFile."""

    pass


class SpectrumFileParsingError(Exception):
    """Failed while parsing a spectrum file."""

    pass


class SpectrumFile(object):
    """Spectrum file parser base class.

    Just instantiate a class with a filename:
        spec = SpectrumFile(filename)

    Then the data are in
        spec.data [counts]
        spec.channels
        spec.energies
        spec.bin_edges_kev
        spec.energy_bin_widths
        spec.energy_bin_edges (deprecated)

    """

    def __init__(self, filename):
        """Initialize the spectrum."""
        self.filename = filename
        assert os.path.exists(self.filename)
        # fields to read from file
        self.spectrum_id = ''
        self.sample_description = ''
        self.detector_description = ''
        self.location_description = ''
        self.hardware_status = ''
        self.collection_start = None
        self.collection_stop = None
        self.realtime = 0.0
        self.livetime = 0.0
        self.num_channels = 0
        # miscellaneous metadata
        self.metadata = {}
        # arrays to be read from file
        self.channels = np.array([], dtype=np.float)
        self.data = np.array([], dtype=np.float)
        self.cal_coeff = []
        # arrays to be calculated using calibration
        self.energies = np.array([], dtype=np.float)
        self.bin_edges_kev = None

    def __str__(self):
        """String form of the spectrum."""

        print_channels = False

        s = ''
        s += 'Filename:              {:s}\n'.format(self.filename)
        s += 'Spectrum ID:           {:s}\n'.format(self.spectrum_id)
        s += 'Sample description:    {:s}\n'.format(self.sample_description)
        s += 'Detector description:  {:s}\n'.format(self.detector_description)
        s += 'Location Description:  {:s}\n'.format(self.location_description)
        s += 'Hardware Status:       {:s}\n'.format(self.hardware_status)
        if self.collection_start is not None:
            s += 'Collection Start:      {:%Y-%m-%d %H:%M:%S}\n'.format(
                self.collection_start)
        else:
            s += 'Collection Start:      None\n'
        if self.collection_stop is not None:
            s += 'Collection Stop:       {:%Y-%m-%d %H:%M:%S}\n'.format(
                self.collection_stop)
        else:
            s += 'Collection Stop:       None\n'
        s += 'Livetime:              {:.2f} sec\n'.format(self.livetime)
        s += 'Realtime:              {:.2f} sec\n'.format(self.realtime)
        if len(self.metadata.keys()) > 0:
            s += 'Metadata:\n'
            for key, value in self.metadata.items():
                s += '    {} : {}\n'.format(key, value)
        s += 'Number of channels:    {:d}\n'.format(self.num_channels)
        if len(self.cal_coeff) > 0:
            s += 'Calibration coeffs:    '
            s += ' '.join(['{:E}'.format(x) for x in self.cal_coeff])
            s += '\n'
        s += 'Data:                  \n'
        if print_channels:
            for ch, dt in zip(self.channels, self.data):
                s += '    {:5.0f}    {:5.0f}\n'.format(ch, dt)
        else:
            s += '    [length {}]\n'.format(len(self.data))
        return s

    @property
    def energy_bin_edges(self):
        warnings.warn('The use of energy_bin_edges is deprecated, ' +
                      'use bin_edges_kev instead', DeprecationWarning)
        return self.bin_edges_kev

    @property
    def energy_bin_widths(self):
        """Retrieve the calibrated width of all the bins."""
        return self.bin_width(self.channels)

    def read(self, verbose=False):
        """Read in the file."""
        raise NotImplementedError('read method not implemented')

    def write(self, filename):
        """Write back to a file."""
        raise NotImplementedError('write method not implemented')

    def apply_calibration(self):
        """Calculate energies corresponding to channels."""
        self.energies = self.channel_to_energy(self.channels)
        n_edges = len(self.energies) + 1
        channel_edges = np.linspace(-0.5, self.channels[-1] + 0.5, num=n_edges)
        self.bin_edges_kev = self.channel_to_energy(channel_edges)

        # check that calibration makes sense, remove calibration if not
        if np.any(np.diff(self.energies) <= 0):
            warnings.warn(
                'Spectrum will be initated without an energy calibration;' +
                'invalid calibration, energies not monotonically increasing.',
                SpectrumFileParsingWarning)
            self.bin_edges_kev = None

    def channel_to_energy(self, channel):
        """Apply energy calibration to the given channel(s)."""
        chan = np.array(channel, dtype=float)
        en = np.zeros_like(chan)
        for j in range(len(self.cal_coeff)):
            en += self.cal_coeff[j] * pow(chan, j)
        return en

    def energy_to_channel(self, energy):
        """Invert the energy calibration to find the channel(s)."""
        energy = np.array(energy, dtype=float)
        return interp1d(self.energies, self.channels)(energy)

    def bin_width(self, channel):
        """Calculate the width of the bin in keV at the channel(s)."""
        en0 = self.channel_to_energy(channel - 0.5)
        en1 = self.channel_to_energy(channel + 0.5)
        return en1 - en0



"""A simple class for caching a pandas DataFrame."""

from __future__ import print_function
import os
import pandas as pd


class CacheError(Exception):
    """Problem fetching, saving, or retrieving cached data."""

    pass


class DataFrameCache(object):
    """Abstract base class for downloading, saving, and retrieving a DataFrame.

    Abstract methods:
      fetch: retrieve the DataFrame from a source.

    Properties (read-only):
      name: the name of the cache
      path: the path where the cache will be stored
      filename: the filename of the cache
      df: the DataFrame itself
      loaded: boolean telling whether the cache has been loaded

    Methods:
      check_path: check for a valid path
      check_file: check that the file exists
      write_file: write the DataFrame to file
      read_file: read the DataFrame from file
      delete_file: delete the cache file
      load: load the DataFrame, from file if available or from fetch()
    """

    name = 'base'
    path = None

    def __init__(self):
        """Initialize the cache.

        Raises:
          CacheError: if the path is invalid
        """

        if self.path is None:
            self.path = os.path.split(__file__)[0]
        self.check_path()
        self.filename = os.path.join(
            self.path, '__df_cache__' + self.name + '.csv')
        self.df = None
        self.loaded = False

    def check_path(self):
        """Test that the path exists.

        Raises:
          CacheError: if the path does not exist.
        """

        if not os.path.exists(self.path):
            raise CacheError('Cache path does not exist: {}'.format(self.path))
        if not os.path.isdir(self.path):
            raise CacheError(
                'Cache path is not a directory: {}'.format(self.path))

    def check_file(self):
        """Test that the file exists.

        Raises:
          CacheError: if the file does not exist.
        """

        if not os.path.exists(self.filename):
            raise CacheError(
                'Cache filename does not exist: {}'.format(self.filename))
        if not os.path.isfile(self.filename):
            raise CacheError(
                'Cache filename is not a file: {}'.format(self.filename))

    def write_file(self):
        """Write the DataFrame to the cache file.

        Raises:
          CacheError: if there was a problem writing the cache to file.
        """

        self.check_path()
        if not self.loaded:
            raise CacheError('Cache has not been fetched or loaded')
        try:
            self.df.to_csv(self.filename, float_format='%.12f')
        except:
            raise CacheError(
                'Problem writing cache to file {}'.format(self.filename))
        self.check_file()

    def read_file(self):
        """Read the cached DataFrame from file.

        Raises:
          CacheError: if there was a problem reading the cache from file.
        """

        self.check_file()
        try:
            self.df = pd.read_csv(self.filename)
        except:
            raise CacheError(
                'Problem reading cache from file {}'.format(self.filename))
        self.loaded = True

    def delete_file(self):
        """Delete the cache file.

        Raises:
          CacheError: if there was a problem deleting the file.
        """

        self.check_file()
        try:
            os.remove(self.filename)
        except:
            raise CacheError(
                'Problem deleting cache file {}'.format(self.filename))
        try:
            self.check_file()
        except CacheError:
            pass  # this should be raised
        else:
            raise CacheError(
                'Cache file was not deleted: {}'.format(self.filename))

    def fetch(self):
        """Fetch the DataFrame to be cached.

        This method must be implemented in child classes.
        """

        raise NotImplementedError('Must implement fetch method')

    def load(self):
        """Read or download the cached DataFrame.

        Raises:
          CacheError: if there was a problem fetching the data or reading
          the cache from file, or if there was a problem writing the cache
          to file.
        """

        try:
            self.read_file()
        except CacheError:
            try:
                self.fetch()
            except CacheError:
                raise CacheError('Cannot read or download DataFrame')
            self.write_file()
            self.read_file()
        self.loaded = True



"""Element symbols, names, Z values, and masses."""

from __future__ import print_function


# Element data (Z, symbol, name, and relative atomic mass) are from Wikipedia:
# https://en.wikipedia.org/wiki/List_of_chemical_elements

_Z_SYMBOL_NAME_MASS = (
    (1, 'H', 'Hydrogen', 1.01),
    (2, 'He', 'Helium', 4.00),
    (3, 'Li', 'Lithium', 6.94),
    (4, 'Be', 'Beryllium', 9.01),
    (5, 'B', 'Boron', 10.81),
    (6, 'C', 'Carbon', 12.01),
    (7, 'N', 'Nitrogen', 14.01),
    (8, 'O', 'Oxygen', 16.00),
    (9, 'F', 'Fluorine', 19.00),
    (10, 'Ne', 'Neon', 20.18),
    (11, 'Na', 'Sodium', 22.99),
    (12, 'Mg', 'Magnesium', 24.30),
    (13, 'Al', 'Aluminum', 26.98),
    (14, 'Si', 'Silicon', 28.09),
    (15, 'P', 'Phosphorus', 30.97),
    (16, 'S', 'Sulfur', 32.06),
    (17, 'Cl', 'Chlorine', 35.45),
    (18, 'Ar', 'Argon', 39.95),
    (19, 'K', 'Potassium', 39.10),
    (20, 'Ca', 'Calcium', 40.08),
    (21, 'Sc', 'Scandium', 44.96),
    (22, 'Ti', 'Titanium', 47.87),
    (23, 'V', 'Vanadium', 50.94),
    (24, 'Cr', 'Chromium', 52.00),
    (25, 'Mn', 'Manganese', 54.94),
    (26, 'Fe', 'Iron', 55.84),
    (27, 'Co', 'Cobalt', 58.93),
    (28, 'Ni', 'Nickel', 58.69),
    (29, 'Cu', 'Copper', 63.55),
    (30, 'Zn', 'Zinc', 65.38),
    (31, 'Ga', 'Gallium', 69.72),
    (32, 'Ge', 'Germanium', 72.63),
    (33, 'As', 'Arsenic', 74.92),
    (34, 'Se', 'Selenium', 78.97),
    (35, 'Br', 'Bromine', 79.90),
    (36, 'Kr', 'Krypton', 83.80),
    (37, 'Rb', 'Rubidium', 85.47),
    (38, 'Sr', 'Strontium', 87.62),
    (39, 'Y', 'Yttrium', 88.91),
    (40, 'Zr', 'Zirconium', 91.22),
    (41, 'Nb', 'Niobium', 92.91),
    (42, 'Mo', 'Molybdenum', 95.95),
    (43, 'Tc', 'Technetium', 98),
    (44, 'Ru', 'Ruthenium', 101.07),
    (45, 'Rh', 'Rhodium', 102.91),
    (46, 'Pd', 'Palladium', 106.42),
    (47, 'Ag', 'Silver', 107.87),
    (48, 'Cd', 'Cadmium', 112.41),
    (49, 'In', 'Indium', 114.82),
    (50, 'Sn', 'Tin', 118.71),
    (51, 'Sb', 'Antimony', 121.76),
    (52, 'Te', 'Tellurium', 127.60),
    (53, 'I', 'Iodine', 126.90),
    (54, 'Xe', 'Xenon', 131.29),
    (55, 'Cs', 'Cesium', 132.91),
    (56, 'Ba', 'Barium', 137.33),
    (57, 'La', 'Lanthanum', 138.91),
    (58, 'Ce', 'Cerium', 140.12),
    (59, 'Pr', 'Praseodymium', 140.91),
    (60, 'Nd', 'Neodymium', 144.24),
    (61, 'Pm', 'Promethium', 145),
    (62, 'Sm', 'Samarium', 150.36),
    (63, 'Eu', 'Europium', 151.96),
    (64, 'Gd', 'Gadolinium', 157.25),
    (65, 'Tb', 'Terbium', 158.93),
    (66, 'Dy', 'Dysprosium', 162.50),
    (67, 'Ho', 'Holmium', 164.93),
    (68, 'Er', 'Erbium', 167.26),
    (69, 'Tm', 'Thulium', 168.93),
    (70, 'Yb', 'Ytterbium', 173.04),
    (71, 'Lu', 'Lutetium', 174.97),
    (72, 'Hf', 'Hafnium', 178.49),
    (73, 'Ta', 'Tantalum', 180.95),
    (74, 'W', 'Tungsten', 183.84),
    (75, 'Re', 'Rhenium', 186.21),
    (76, 'Os', 'Osmium', 190.23),
    (77, 'Ir', 'Iridium', 192.22),
    (78, 'Pt', 'Platinum', 195.08),
    (79, 'Au', 'Gold', 196.97),
    (80, 'Hg', 'Mercury', 200.59),
    (81, 'Tl', 'Thallium', 204.38),
    (82, 'Pb', 'Lead', 207.20),
    (83, 'Bi', 'Bismuth', 208.98),
    (84, 'Po', 'Polonium', 209),
    (85, 'At', 'Astatine', 210),
    (86, 'Rn', 'Radon', 222),
    (87, 'Fr', 'Francium', 223),
    (88, 'Ra', 'Radium', 226),
    (89, 'Ac', 'Actinium', 227),
    (90, 'Th', 'Thorium', 232.04),
    (91, 'Pa', 'Protactinium', 231.04),
    (92, 'U', 'Uranium', 238.03),
    (93, 'Np', 'Neptunium', 237),
    (94, 'Pu', 'Plutonium', 244),
    (95, 'Am', 'Americium', 243),
    (96, 'Cm', 'Curium', 247),
    (97, 'Bk', 'Berkelium', 247),
    (98, 'Cf', 'Californium', 251),
    (99, 'Es', 'Einsteinium', 252),
    (100, 'Fm', 'Fermium', 257),
    (101, 'Md', 'Mendelevium', 258),
    (102, 'No', 'Nobelium', 259),
    (103, 'Lr', 'Lawrencium', 266),
    (104, 'Rf', 'Rutherfordium', 267),
    (105, 'Db', 'Dubnium', 268),
    (106, 'Sg', 'Seaborgium', 269),
    (107, 'Bh', 'Bohrium', 270),
    (108, 'Hs', 'Hassium', 269),
    (109, 'Mt', 'Meitnerium', 278),
    (110, 'Ds', 'Darmstadtium', 281),
    (111, 'Rg', 'Roentgenium', 282),
    (112, 'Cn', 'Copernicium', 285),
    (113, 'Nh', 'Nihonium', 286),
    (114, 'Fl', 'Flerovium', 289),
    (115, 'Mc', 'Moscovium', 289),
    (116, 'Lv', 'Livermorium', 293),
    (117, 'Ts', 'Tennessine', 294),
    (118, 'Og', 'Oganesson', 294),
)


ZS = set([d[0] for d in _Z_SYMBOL_NAME_MASS])
SYMBOLS = set([d[1] for d in _Z_SYMBOL_NAME_MASS])
SYMBOLS_LOWER = set([d[1].lower() for d in _Z_SYMBOL_NAME_MASS])
SYMBOLS_UPPER = set([d[1].upper() for d in _Z_SYMBOL_NAME_MASS])
NAMES = set([d[2] for d in _Z_SYMBOL_NAME_MASS])
NAMES_LOWER = set([d[2].lower() for d in _Z_SYMBOL_NAME_MASS])
NAMES_UPPER = set([d[2].upper() for d in _Z_SYMBOL_NAME_MASS])

_SYMBOL_FROM_SYMBOL_LOWER = {d[1].lower(): d[1] for d in _Z_SYMBOL_NAME_MASS}
_SYMBOL_FROM_SYMBOL_UPPER = {d[1].upper(): d[1] for d in _Z_SYMBOL_NAME_MASS}

_NAME_FROM_NAME_LOWER = {d[2].lower(): d[2] for d in _Z_SYMBOL_NAME_MASS}
_NAME_FROM_NAME_UPPER = {d[2].upper(): d[2] for d in _Z_SYMBOL_NAME_MASS}

_Z_FROM_SYMBOL = {d[1]: d[0] for d in _Z_SYMBOL_NAME_MASS}
_Z_FROM_NAME = {d[2]: d[0] for d in _Z_SYMBOL_NAME_MASS}

_SYMBOL_FROM_Z = {d[0]: d[1] for d in _Z_SYMBOL_NAME_MASS}
_SYMBOL_FROM_NAME = {d[2]: d[1] for d in _Z_SYMBOL_NAME_MASS}

_NAME_FROM_Z = {d[0]: d[2] for d in _Z_SYMBOL_NAME_MASS}
_NAME_FROM_SYMBOL = {d[1]: d[2] for d in _Z_SYMBOL_NAME_MASS}

_MASS_FROM_Z = {d[0]: d[3] for d in _Z_SYMBOL_NAME_MASS}
_MASS_FROM_SYMBOL = {d[1]: d[3] for d in _Z_SYMBOL_NAME_MASS}


class ElementError(Exception):
    """Problem with element properties."""

    pass


class ElementZError(ElementError):
    """Bad element Z value."""

    pass


class ElementSymbolError(ElementError):
    """Bad element symbol."""

    pass


class ElementNameError(ElementError):
    """Bad element name."""

    pass


def validated_z(z):
    """Convert z into a valid element atomic number.

    Args:
      z: numeric or string type representing the atomic number Z.

    Returns:
      Integer z that is a valid atomic number.

    Raises:
      ElementZError: if z cannot be converted to integer or is out of range.
    """

    try:
        int(z)
    except ValueError:
        raise ElementZError('Element Z="{}" invalid'.format(z))
    if int(z) not in ZS:
        raise ElementZError('Element Z="{}" not found'.format(z))
    return int(z)


def validated_symbol(sym):
    """Convert symbol into a valid mixed-case element symbol.

    Args:
      sym: a string type

    Returns:
      A valid mixed-case element symbol that matches sym.

    Raises:
      ElementSymbolError: if sym does not match a valid symbol.
    """

    try:
        sym.lower()
    except AttributeError:
        raise ElementSymbolError('Element symbol "{}" invalid'.format(sym))
    if sym.lower() not in SYMBOLS_LOWER:
        raise ElementSymbolError('Element symbol "{}" not found'.format(sym))
    return _SYMBOL_FROM_SYMBOL_LOWER[sym.lower()]


def validated_name(nm):
    """Convert name into a valid mixed-case element name.

    Args:
      nm: a string type

    Returns:
      A valid mixed-case element name that matches nm.

    Raises:
      ElementNameError: if nm does not match a valid name.
    """

    try:
        nm.lower()
    except AttributeError:
        raise ElementNameError('Element name "{}" invalid'.format(nm))
    # special case: Alumin[i]um
    if nm.lower() == 'aluminium':
        nm = 'Aluminum'
    # special case: C[a]esium
    if nm.lower() == 'caesium':
        nm = 'Cesium'
    if nm.lower() not in NAMES_LOWER:
        raise ElementNameError('Element name "{}" not found'.format(nm))
    return _NAME_FROM_NAME_LOWER[nm.lower()]


def element_z(sym_or_name):
    """Convert element symbol or name into a valid element atomic number Z.

    Args:
      sym_or_name: string type representing an element symbol or name.

    Returns:
      Integer z that is a valid atomic number matching the symbol or name.

    Raises:
      ElementZError: if the symbol or name cannot be converted.
    """

    try:
        return _Z_FROM_SYMBOL[validated_symbol(sym_or_name)]
    except ElementSymbolError:
        pass
    try:
        return _Z_FROM_NAME[validated_name(sym_or_name)]
    except ElementNameError:
        raise ElementZError('Must supply either the element symbol or name')


def element_symbol(name_or_z):
    """Convert element name or Z into a valid mixed-case element symbol.

    Args:
      name_or_z: string or numeric type representing an element name or Z.

    Returns:
      A valid mixed-case element symbol that matches the name or Z.

    Raises:
      ElementSymbolError: if the symbol or name cannot be converted.
    """

    try:
        return _SYMBOL_FROM_Z[validated_z(name_or_z)]
    except ElementZError:
        pass
    try:
        return _SYMBOL_FROM_NAME[validated_name(name_or_z)]
    except ElementNameError:
        raise ElementSymbolError('Must supply either the Z or element name')


def element_name(sym_or_z):
    """Convert element symbol or Z into a valid mixed-case element name.

    Args:
      sym_or_z: string or numeric type representing an element symbol or Z.

    Returns:
      A valid mixed-case element name that matches the symbol or Z.

    Raises:
      ElementNameError: if the symbol or Z cannot be converted.
    """

    try:
        return _NAME_FROM_SYMBOL[validated_symbol(sym_or_z)]
    except ElementSymbolError:
        pass
    try:
        return _NAME_FROM_Z[validated_z(sym_or_z)]
    except ElementZError:
        raise ElementNameError('Must supply either the element symbol or Z')


class Element(object):
    """Basic properties (symbol, name, Z, and mass) of an element.

    Also provides string formatting:
    >>> elem = Element('Ge')
    >>> '{:%n(%s) Z=%z}'.format(elem)
    'Germanium(Ge) Z=32'

    Properties:
      symbol (read-only): the mixed-case element symbol (e.g., "Ge")
      name (read-only): the mixed-case element name (e.g., "Germanium")
      Z (read-only): an integer giving the atomic number (e.g., 32)
      atomic_mass (read-only): a float giving the atomic mass in amu
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, arg):
        """Instantiate by providing a name, symbol, or Z value.

        Args:
          arg: a numeric or string type giving a name, symbol, or Z value.

        Raises:
          ElementError: if Element could not be instantiated.
        """

        self.symbol = None
        self.name = None
        self.Z = None
        self.atomic_mass = None
        try:
            self._init_sym(arg)
        except ElementSymbolError:
            try:
                self._init_name(arg)
            except ElementNameError:
                try:
                    self._init_z(arg)
                except ElementZError:
                    raise ElementError(
                        'Could not instantiate Element: {}'.format(arg))
        self.atomic_mass = _MASS_FROM_SYMBOL[self.symbol]

    def _init_sym(self, arg):
        """Initialize with an element symbol."""
        self.symbol = validated_symbol(arg)
        self.name = element_name(self.symbol)
        self.Z = element_z(self.symbol)

    def _init_name(self, arg):
        """Initialize with an element name."""
        self.name = validated_name(arg)
        self.symbol = element_symbol(self.name)
        self.Z = element_z(self.name)

    def _init_z(self, arg):
        """Initialize with an element Z."""
        self.Z = validated_z(arg)
        self.symbol = element_symbol(self.Z)
        self.name = element_name(self.Z)

    def __str__(self):
        """Define behavior of str() on this class."""
        return '{}'.format(self)

    def __format__(self, formatstr):
        """Define behavior of string's format method.

        Format codes:
            '%s': element symbol
            '%n': element name
            '%z': element Z
        """

        str0 = str(formatstr)
        if len(str0) == 0:
            str0 = '%n(%s) Z=%z'
        str0 = str0.replace('%s', self.symbol)
        str0 = str0.replace('%n', self.name)
        str0 = str0.replace('%z', '{}'.format(self.Z))
        return str0

    def __eq__(self, other):
        """Define equality of two elements."""
        try:
            return (
                self.name == other.name and self.symbol == other.symbol and
                self.Z == other.Z)
        except:
            raise ElementError('Cannot determine equality')
